AutomatonBEnv: trait
  includes 
    Records, Set(BalRec), Array(I, Nat), Array(I, Bool), IOAutomaton(BEnv),
    Set(OpRec), Array(I, Set[OpRec])
  States[BEnv] tuple of 
    ops: Set[OpRec], bals: Set[BalRec], lastSeqno: Array[I, Nat],
    chosenOps: Set[OpRec], mustInclude: Array[I, Set[OpRec]], 
    active: Array[I, Bool]
  introduces
    OK, requestBalance, doBalance: I -> Actions[BEnv]
    reportBalance, requestDeposit, requestWithdrawal: Int, I -> Actions[BEnv]
    almostsame: States[BEnv], States[BEnv], I -> Bool
  asserts 
   sort Actions[BEnv] generated freely by
        OK, reportBalance, requestDeposit, requestWithdrawal, requestBalance,
        doBalance
        % This does not take account of the "where n > 0" clause

   with i, j: I, n: Int, s, s': States[BEnv], x, y: OpRec
     isOutput(OK(i));
     isOutput(reportBalance(n, i));
     isOutput(requestDeposit(n, i));
     isOutput(requestWithdrawal(n, i));
     isInput(requestBalance(i));
     isInternal(doBalance(i));

     almostsame(s, s', i) <=> \A j (j ~= i => s'.active[j] = s.active[j]);

     start(s) <=> s.ops = {} /\ s.bals = {} /\ s.lastSeqno = constant(0)
                  /\ s.mustInclude = constant({}) 
                  /\ s.active = constant(false);

     enabled(s, requestDeposit(n, i)) 
       <=> ~s.active[i];
     effect(s, requestDeposit(n, i), s') 
       <=> (s'.lastSeqno)[i] = (s.lastSeqno)[i] + 1
           /\ \A j (i ~= j => s'.lastSeqno[j] = s.lastSeqno[j])
           /\ s'.ops = insert([n, i, nat2pos(s'.lastSeqno[i]), false], s.ops)
           /\ s'.bals = s.bals
           /\ s'.chosenOps = s.chosenOps
           /\ s'.mustInclude = s.mustInclude
           /\ s'.active[i] /\ almostsame(s, s', i);

     enabled(s, requestWithdrawal(n, i)) 
       <=> ~s.active[i];
     effect(s, requestWithdrawal(n, i), s') 
       <=> s'.lastSeqno[i] = s.lastSeqno[i] + 1
           /\ \A j (i ~= j => s'.lastSeqno[j] = s.lastSeqno[j])
           /\ s'.ops = insert([-n, i, nat2pos(s'.lastSeqno[i]), false], s.ops)
           /\ s'.bals = s.bals
           /\ s'.chosenOps = s.chosenOps
           /\ s'.mustInclude = s.mustInclude
           /\ s'.active[i] /\ almostsame(s, s', i);

     enabled(s, requestBalance(i)) 
       <=> ~s.active[i];
     effect(s, requestBalance(i), s') 
       <=> s'.bals = insert([i, null], s.bals)
           /\ \A x:OpRec (x \in s'.mustInclude[i] 
                            <=> x \in s.ops /\ x.reported)
           /\ s'.lastSeqno = s.lastSeqno
           /\ s'.ops = s.ops
           /\ s'.chosenOps = s.chosenOps
           /\ s'.active[i] /\ almostsame(s, s', i);

     enabled(s, OK(i))
       <=> \E x: OpRec (x \in s.ops /\ x.location = i /\ ~x.reported);
     effect(s, OK(i), s') 
       <=> \E x: OpRec (
                 x \in s.ops /\ x.location = i /\ ~x.reported
              /\ s'.ops = insert(set_reported(x, true), delete(x, s.ops)))
           /\ s'.lastSeqno = s.lastSeqno
           /\ s'.bals = s.bals
           /\ s'.chosenOps = s.chosenOps
           /\ s'.mustInclude = s.mustInclude
           /\ s'.active[i] = false /\ almostsame(s, s', i);

     enabled(s, doBalance(i))
       <=> [i, null] \in s.bals;
     effect(s, doBalance(i), s') 
       <=>    \A y: OpRec (y.location = i /\ y \in s.ops => y \in s'.chosenOps)
           /\ s'.chosenOps \subseteq s.ops
           /\ s.mustInclude[i] \subseteq s'.chosenOps
           /\ s'.bals = insert([i, define(totalAmount(s'.chosenOps))], 
                               delete([i, null], s.bals))
           /\ s'.lastSeqno = s.lastSeqno
           /\ s'.ops = s.ops
           /\ s'.active = s.active;
              
     enabled(s, reportBalance(n, i))
       <=> [i, define(n)] \in s.bals      ;
     effect(s, reportBalance(n, i), s')
       <=> s'.bals = delete([i, define(n)], s.bals)
           /\ s'.lastSeqno = s.lastSeqno
           /\ s'.ops = s.ops
           /\ s'.chosenOps = s.chosenOps
           /\ s'.mustInclude = s.mustInclude
           /\ s'.active[i] = false /\ almostsame(s, s', i);
