AutomatonCEnv: trait
  includes 
    Records, Set(BalRec1), Set(Pos), Set(OpRec1), Sequence(Msg), 
    Array(I, Bool), Array(I, Set[OpRec1]), Array2(I, I, Seq[Msg]), 
    Array(I, States[C]), IOAutomaton(CEnv)
  States[C] tuple of
    ops: Set[OpRec1], reports: Set[Pos], bals: Set[BalRec1], 
    lastSeqno: Nat, reqSent, respRcvd, reqRcvd: Array[I, Bool]
  States[CEnv] tuple of 
    C: Array[I, States[C]], 
    active: Array[I, Bool], 
    mustInclude: Array[I, Set[OpRec1]],
    queue: Array[I, I, Seq[Msg]]      % Ignoring for now i ~= j
  Msg union of set: Set[OpRec1], req: String
  introduces
    OK, requestBalance, doBalance: I -> Actions[CEnv]
    reportBalance, requestDeposit, requestWithdrawal: Int, I -> Actions[CEnv]
    send, receive: Msg, I, I -> Actions[CEnv]
    updateSeqno: States[CEnv], States[CEnv], I -> Bool
    updateActive: States[CEnv], States[CEnv], I, Bool -> Bool
  asserts 
   sort Actions[CEnv] generated freely by
     OK, reportBalance, requestDeposit, requestWithdrawal, requestBalance,
     doBalance, receive, send
     % This does not take account of the "where n > 0" clause

   with i, j, k, i1, j1: I, n: Int, s, s': States[CEnv], x, y: OpRec1, 
        b: Bool, m: Set[OpRec1], m: Msg, x: String
     isOutput(OK(i));
     isOutput(reportBalance(n, i));
     isOutput(requestDeposit(n, i));
     isOutput(requestWithdrawal(n, i));
     isOutput(requestBalance(i));
     isInternal(doBalance(i));
     isInternal(send(m, i, j));
     isInternal(receive(m, i, j));

     updateSeqno(s, s', i) <=>
          (s'.C[i]).lastSeqno = (s.C[i]).lastSeqno + 1
       /\ \A j (j ~= i => (s'.C[j]).lastSeqno = (s.C[j]).lastSeqno);

     updateActive(s, s', i, b) <=>
       s'.active[i] = b /\ \A j (j ~= i => s'.active[j] = s.active[j]);

     start(s) <=> s.C = constant([{}, {}, {}, 0, constant(false),
                                  constant(false), constant(false)])
                  /\ s.active = constant(false)
                  /\ s.mustInclude = constant({}) 
                  /\ s.queue = constant({});

     enabled(s, requestDeposit(n, i)) 
       <=> ~s.active[i];
     effect(s, requestDeposit(n, i), s') 
       <=>    updateSeqno(s, s', i)
           /\ (s'.C[i]).ops 
                = insert([n, i, nat2pos((s'.C[i]).lastSeqno)], (s.C[i]).ops)
           /\ \A j (i ~= j => (s'.C[j]).ops = (s.C[j]).ops)
           /\ \A k ((s'.C[k]).bals = (s.C[k]).bals)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).reqSent = ((s.C)[k]).reqSent)
           /\ \A k (((s'.C)[k]).reqRcvd = ((s.C)[k]).reqRcvd)
           /\ \A k (((s'.C)[k]).respRcvd = ((s.C)[k]).respRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue = s.queue
           /\ updateActive(s, s', i, true);

     enabled(s, requestWithdrawal(n, i)) 
       <=> ~s.active[i];
     effect(s, requestWithdrawal(n, i), s') 
       <=>    updateSeqno(s, s', i)
           /\ (s'.C[i]).ops 
                = insert([-n, i, nat2pos((s'.C[i]).lastSeqno)], (s.C[i]).ops)
           /\ \A j (i ~= j => (s'.C[j]).ops = (s.C[j]).ops)
           /\ \A k ((s'.C[k]).bals = (s.C[k]).bals)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).reqSent = ((s.C)[k]).reqSent)
           /\ \A k (((s'.C)[k]).reqRcvd = ((s.C)[k]).reqRcvd)
           /\ \A k (((s'.C)[k]).respRcvd = ((s.C)[k]).respRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue = s.queue
           /\ updateActive(s, s', i, true);

     enabled(s, requestBalance(i)) 
       <=> ~s.active[i];
     effect(s, requestBalance(i), s') 
       <=>    (s'.C[i]).bals = insert([null], (s.C[i]).bals)
           /\ \A j (i ~= j => (s'.C[j]).bals = (s.C[j]).bals)
           /\ (s'.C[i]).reqSent = constant(false)
           /\ (s'.C[i]).respRcvd = constant(false)
           /\ \A j (i ~= j => (s'.C[j]).reqSent = (s.C[j]).reqSent
                               /\ (s'.C[j]).respRcvd = (s.C[j]).respRcvd)
           /\ \A k (((s'.C)[k]).ops = ((s.C)[k]).ops)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).lastSeqno = ((s.C)[k]).lastSeqno)
           /\ \A k (((s'.C)[k]).reqRcvd = ((s.C)[k]).reqRcvd)
           /\ \A x:OpRec1 (x \in s'.mustInclude[i] <=>
                           \E j (x \in (s.C[j]).ops /\ x.location = j
                                 /\ x.seqno \in (s.C[j]).reports))
           /\ \A j (j ~= i => s'.mustInclude[j] = s.mustInclude[j])
           /\ s'.queue = s.queue
           /\ updateActive(s, s', i, true);

     enabled(s, OK(i))
       <=> \E x: OpRec1 (x \in (s.C[i]).ops /\ x.location = i 
                         /\ ~((x.seqno) \in (s.C[i]).reports));
     effect(s, OK(i), s') 
       <=> \E x: OpRec1 (x \in (s.C[i]).ops /\ x.location = i
                          /\ ~((x.seqno) \in (s.C[i]).reports)
                          /\ (s'.C[i]).reports = insert(x.seqno, (s.C[i]).reports))
           /\ \A j (i ~= j => (s'.C[j]).reports = (s.C[j]).reports)
           /\ \A k (((s'.C)[k]).ops = ((s.C)[k]).ops)
           /\ \A k ((s'.C[k]).bals = (s.C[k]).bals)
           /\ \A k (((s'.C)[k]).lastSeqno = ((s.C)[k]).lastSeqno)
           /\ \A k (((s'.C)[k]).reqSent = ((s.C)[k]).reqSent)
           /\ \A k (((s'.C)[k]).reqRcvd = ((s.C)[k]).reqRcvd)
           /\ \A k (((s'.C)[k]).respRcvd = ((s.C)[k]).respRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue = s.queue
           /\ updateActive(s, s', i, false);

     enabled(s, doBalance(i))
       <=> [null] \in (s.C[i]).bals /\ \A j (j ~= i => (s.C[i]).respRcvd[j]);
     effect(s, doBalance(i), s') 
       <=>    (s'.C[i]).bals = insert([define(totalAmount((s.C[i]).ops))], 
                                    delete([null], (s.C[i]).bals))
           /\ \A j (i ~= j => (s'.C[j]).bals = (s.C[j]).bals)
           /\ \A k (((s'.C)[k]).ops = ((s.C)[k]).ops)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).lastSeqno = ((s.C)[k]).lastSeqno)
           /\ \A k (((s'.C)[k]).reqSent = ((s.C)[k]).reqSent)
           /\ \A k (((s'.C)[k]).reqRcvd = ((s.C)[k]).reqRcvd)
           /\ \A k (((s'.C)[k]).respRcvd = ((s.C)[k]).respRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue = s.queue
           /\ s'.active = s.active;
              
     enabled(s, reportBalance(n, i))
       <=> [define(n)] \in (s.C[i]).bals;
     effect(s, reportBalance(n, i), s')
       <=>    (s'.C[i]).bals = delete([define(n)], (s.C[i]).bals)
           /\ \A j (i ~= j => (s'.C[j]).bals = (s.C[j]).bals)
           /\ \A k (((s'.C)[k]).ops = ((s.C)[k]).ops)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).lastSeqno = ((s.C)[k]).lastSeqno)
           /\ \A k (((s'.C)[k]).reqSent = ((s.C)[k]).reqSent)
           /\ \A k (((s'.C)[k]).reqRcvd = ((s.C)[k]).reqRcvd)
           /\ \A k (((s'.C)[k]).respRcvd = ((s.C)[k]).respRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue = s.queue
           /\ updateActive(s, s', i, false);

     enabled(s, send(req(x), i, j))
       <=> i ~= j /\ ~(s.C[i]).reqSent[j] /\ [null] \in (s.C[i]).bals;
     effect(s, send(req(x), i, j), s') 
       <=>    (s'.C[i]).reqSent[j]
           /\ \A k (k ~= i => (s'.C[k]).reqSent = (s.C[k]).reqSent)
           /\ \A k (k ~= j => (s'.C[i]).reqSent[k] = (s.C[i]).reqSent[k])
           /\ \A k (((s'.C)[k]).ops = ((s.C)[k]).ops)
           /\ \A k ((s'.C[k]).bals = (s.C[k]).bals)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).lastSeqno = ((s.C)[k]).lastSeqno)
           /\ \A k (((s'.C)[k]).reqRcvd = ((s.C)[k]).reqRcvd)
           /\ \A k (((s'.C)[k]).respRcvd = ((s.C)[k]).respRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue[i, j] = s.queue[i, j] |- req(x)
           /\ \A i1 \A j1 (i ~= i1 \/ j ~= j1 
                             => s'.queue[i1, j1] = s.queue[i1, j1])
           /\ s'.active = s.active;

     enabled(s, receive(req(x), j, i))
       <=> i ~= j /\ s.queue[j, i] ~= {} /\ req(x) = head(s.queue[j, i]);
     effect(s, receive(req(x), j, i), s') 
       <=>    (s'.C[i]).reqRcvd[j]
           /\ \A k (k ~= i => (s'.C[k]).reqRcvd = (s.C[k]).reqRcvd)
           /\ \A k (k ~= j => (s'.C[i]).reqRcvd[k] = (s.C[i]).reqRcvd[k])
           /\ \A k (((s'.C)[k]).ops = ((s.C)[k]).ops)
           /\ \A k ((s'.C[k]).bals = (s.C[k]).bals)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).lastSeqno = ((s.C)[k]).lastSeqno)
           /\ \A k (((s'.C)[k]).reqSent = ((s.C)[k]).reqSent)
           /\ \A k (((s'.C)[k]).respRcvd = ((s.C)[k]).respRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue[j, i] = tail(s.queue[j, i])
           /\ \A i1 \A j1 (i ~= i1 \/ j ~= j1 
                             => s'.queue[j1, i1] = s.queue[j1, i1])
           /\ s'.active = s.active;

     enabled(s, send(set(m), i, j))
       <=>  i ~= j /\ m = (s.C[i]).ops /\ (s.C[i]).reqRcvd[j];
     effect(s, send(set(m), i, j), s') 
       <=>   ~(s'.C[i]).reqRcvd[j]
           /\ \A k (k ~= i => (s'.C[k]).reqRcvd = (s.C[k]).reqRcvd)
           /\ \A k (k ~= j => (s'.C[i]).reqRcvd[k] = (s.C[i]).reqRcvd[k])
           /\ \A k (((s'.C)[k]).ops = ((s.C)[k]).ops)
           /\ \A k ((s'.C[k]).bals = (s.C[k]).bals)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).lastSeqno = ((s.C)[k]).lastSeqno)
           /\ \A k (((s'.C)[k]).reqSent = ((s.C)[k]).reqSent)
           /\ \A k (((s'.C)[k]).respRcvd = ((s.C)[k]).respRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue[i, j] = s.queue[i, j] |- set(m)
           /\ \A i1 \A j1 (i ~= i1 \/ j ~= j1 
                             => s'.queue[i1, j1] = s.queue[i1, j1])
           /\ s'.active = s.active;

     enabled(s, receive(set(m), j, i))
       <=> i ~= j /\ s.queue[j, i] ~= {} /\ set(m) = head(s.queue[j, i]);
     effect(s, receive(set(m), j, i), s') 
       <=>    (s'.C[i]).respRcvd[j]
           /\ \A k (k ~= i => (s'.C[k]).respRcvd = (s.C[k]).respRcvd)
           /\ \A k (k ~= j => (s'.C[i]).respRcvd[k] = (s.C[i]).respRcvd[k])
           /\ (s'.C[i]).ops = (s.C[i]).ops \U m
           /\ \A j (j ~= i => (s'.C[j]).ops = (s.C[j]).ops)
           /\ \A k ((s'.C[k]).bals = (s.C[k]).bals)
           /\ \A k (((s'.C)[k]).reports = ((s.C)[k]).reports)
           /\ \A k (((s'.C)[k]).lastSeqno = ((s.C)[k]).lastSeqno)
           /\ \A k (((s'.C)[k]).reqSent = ((s.C)[k]).reqSent)
           /\ \A k (((s'.C)[k]).reqRcvd = ((s.C)[k]).reqRcvd)
           /\ s'.mustInclude = s.mustInclude
           /\ s'.queue[j, i] = tail(s.queue[j, i])
           /\ \A i1 \A j1 (i ~= i1 \/ j ~= j1 
                             => s'.queue[j1, i1] = s.queue[j1, i1])
           /\ s'.active = s.active;
