% This specification is for documentation only.  It is not used in any proof.
% See AutomatonAEnv.lsl instead.

AutomatonA: trait
  includes Records, Set(BalRec), Array(I, Nat), IOAutomaton(A)
  introduces
    OK, requestBalance, doBalance: I         -> Actions[A]
    reportBalance, requestDeposit,
      requestWithdrawal:           Int, I    -> Actions[A]
    __.ops:                        States[A] -> Set[OpRec]
    __.bals:                       States[A] -> Set[BalRec]
    __.lastSeqno:                  States[A] -> Array[I, Nat]
    __.chosenOps:                  States[A] -> Set[OpRec]
  asserts with i, j: I, n: Int, s, s': States[A], x, y: OpRec
   sort Actions[A] generated freely by
        OK, reportBalance, requestDeposit, requestWithdrawal, requestBalance,
        doBalance;
        % This does not take account of the "where n > 0" clause

     isOutput(OK(i));
     isOutput(reportBalance(n, i));
     isInput(requestDeposit(n, i));
     isInput(requestWithdrawal(n, i));
     isInternal(doBalance(i));

     start(s) <=> s.ops = {} /\ s.bals = {} /\ s.lastSeqno = constant(0);

     effect(s, requestDeposit(n, i), s') 
       <=> (s'.lastSeqno)[i] = (s.lastSeqno)[i] + 1
           /\ \A j (i ~= j => s'.lastSeqno[j] = s.lastSeqno[j])
           /\ s'.ops = insert([n, i, pos(s'.lastSeqno[i]), false], s.ops)
           /\ s'.bals = s.bals
           /\ s'.chosenOps = s.chosenOps;

     effect(s, requestWithdrawal(n, i), s') 
       <=> s'.lastSeqno[i] = s.lastSeqno[i] + 1
           /\ \A j (i ~= j => s'.lastSeqno[j] = s.lastSeqno[j])
           /\ s'.ops = insert([-n, i, pos(s'.lastSeqno[i]), false], s.ops)
           /\ s'.bals = s.bals
           /\ s'.chosenOps = s.chosenOps;

     effect(s, requestBalance(i), s') 
       <=> s'.bals = insert([i, null], s.bals)
           /\ s'.lastSeqno = s.lastSeqno
           /\ s'.ops = s.ops
           /\ s'.chosenOps = s.chosenOps;

     enabled(s, OK(i))
       <=> \E x: OpRec (x \in s.ops /\ x.location = i /\ ~x.reported);
     effect(s, OK(i), s') 
       <=> \E x: OpRec (
                 x \in s.ops /\ x.location = i /\ ~x.reported
              /\ s'.ops = insert(set_reported(x, true), delete(x, s.ops)))
           /\ s'.lastSeqno = s.lastSeqno
           /\ s'.bals = s.bals
           /\ s'.chosenOps = s.chosenOps;

     enabled(s, doBalance(i))
       <=> [i, null] \in s.bals;
     effect(s, doBalance(i), s') 
       <=>    \A y: OpRec (y.location = i  /\ y \in s.ops 
                              => y \in s'.chosenOps)
           /\ s'.chosenOps \subseteq s.ops
           /\ s'.bals = insert([i, define(totalAmount(s'.chosenOps))], 
                               delete([i, null], s.bals))
           /\ s'.lastSeqno = s.lastSeqno
           /\ s'.ops = s.ops;
              
     enabled(s, reportBalance(n, i))
       <=> [i, define(n)] \in s.bals      ;
     effect(s, reportBalance(n, i), s')
       <=> s'.bals = delete([i, define(n)], s.bals)
           /\ s'.lastSeqno = s.lastSeqno
           /\ s'.ops = s.ops
           /\ s'.chosenOps = s.chosenOps;
