theory AuxDist = Main + Set + Array:

theorem existenceSome: "(EX a. (f a)) ==> f (SOME a. (f a))"
  apply (auto)
  apply (simp add: someI)
done



types Node = nat
datatype Value = makeValue nat

types quorum = "(nat set)"

datatype Ballot = 
  createBallot nat Node

consts
  procid :: "Ballot => Node" 
primrec
  procid_def [simp]:
  "procid (createBallot seq n) = n"


consts
  seqno :: "Ballot => nat"
primrec
  seqno_def [simp]:
  "seqno (createBallot seq n) = seq"

consts
 compareBallots :: "Ballot => Ballot => bool"

defs compareBallots_def:
"compareBallots b1 b2 == 
  if ((seqno b1) = (seqno b2))
  then (procid b1) < (procid b2)
  else (seqno b1) < (seqno b2)
"

syntax
  "compareBallots" :: "[Ballot, Ballot] => bool" (infixl "<" 65)

theorem seqno_prop [simp]: "seqno (createBallot seq p) == seq"
  apply (simp)
done

theorem procid_prop [simp]: "procid (createBallot seq p) == p"
  apply (simp)
done

theorem compareBallots_test: "seq1 < seq2 ==> (createBallot seq1 (p1::Node)) < (createBallot seq2 (p2::Node))"
  apply (simp add: compareBallots_def)
done

theorem compareBallots_test2: "seq1 = seq2 & p1 < p2 ==> (createBallot seq1 (p1::Node)) < (createBallot seq2 (p2::Node))"
  apply (simp add: compareBallots_def)
done



constdefs
 rquorums :: "(Node set) set"
 "rquorums == SOME s . ((s ~= {}) & (ALL s' : s . (s' ~= {})))"

constdefs
 wquorums :: "(Node set) set"
 "wquorums == {s. ALL s' .
  (s' : rquorums) -->
  (EX n . (n : s & n : s'))
}"

theorem quorumIntersect:
  assumes p0: "wq : wquorums"
      and p1: "rq : rquorums"
  shows "EX n . (n : wq & n : rq)"
proof -
  have f0: "ALL s' .
  (s' : rquorums) -->
  (EX n . (n : wq & n : s'))"
    apply (insert wquorums_def p0)
    apply (auto)  (* WHY DOESN"T SIMP WORK!?! *)
  done
  show "?thesis"
    apply (insert f0 p1)
    apply (auto)
  done
qed

constdefs 
 getIntersect :: "Node set => Node set => Node"
 "getIntersect wq rq ==
   if (wq : wquorums & rq : rquorums)
   then SOME n . (n : wq & n : rq) 
   else SOME n . False"


theorem getIntersect_good:
  assumes p0: "wq : wquorums"
      and p1: "rq : rquorums"
  shows "(getIntersect wq rq) : wq & (getIntersect wq rq) : rq"
  apply (simp add: getIntersect_def p0 p1)
  apply (auto)
proof -
  have f0: "EX n . (n : wq & n : rq)"
    by (simp add: p0 p1 quorumIntersect)
  have f1: "(SOME n. n : wq & n : rq) : wq & (SOME n. n : wq & n : rq) : rq"
    apply (rule existenceSome)
    apply (simp add: f0)
  done
  show "(SOME n. n : wq & n : rq) : wq"
    by (simp add: f1)
  show "(SOME n. n : wq & n : rq) : rq"
    by (simp add: f1)
  qed

constdefs
 dead :: "(Node, (Ballot set)) Array => Ballot set"
 "dead abstained ==
  {b .
   EX rq . (
      (rq : rquorums) &
      (ALL n : rq . b : (sub abstained n))
   )
  }
 "

constdefs
 abstainedQuorum :: 
   "(Node, (Ballot set)) Array => Ballot => Node set"
 "abstainedQuorum abstained b ==
  if (b : (dead abstained))
  then SOME rq . (
                  (rq : rquorums) &
                  (ALL n : rq . b : (sub abstained n))
                 )
  else {}"

theorem abstainedQuorum_good:
  assumes p0: "b : dead abstained"
  defines "rq == abstainedQuorum abstained b"
  shows "(rq : rquorums) &
         (ALL n : rq . b : (sub abstained n))"
proof -
  have f0: "EX rq . (
                   (rq : rquorums) &
                   (ALL n : rq . b : (sub abstained n)))"
    apply (insert dead_def p0)
    apply (blast)
  done
  show "?thesis"
    apply (simp add: abstainedQuorum_def p0 rq_def)
    apply (rule existenceSome)
    apply (insert f0)
    apply (auto)
  done
qed

theorem rquorumsProp:
  "(rquorums ~= {}) & (ALL s' : rquorums . (s' ~= {}))"
  apply (simp add: rquorums_def)
  apply (rule existenceSome)
  apply (rule exI)
proof -
  def a == "{{5}}"
  show "a ~= {} & (ALL s':a. s' ~= {})"
    by (simp add: a_def)
qed

theorem rquorumsExist:
  assumes p0: "rq : rquorums"
    shows "EX n . n : rq"
  proof -
  have f0: "rq ~= {}"
    apply (simp add: rquorumsProp p0)
  done
  show "?thesis"
    apply (insert f0)
    apply (blast)
  done
qed

theorem wquorumsExist:
  assumes p0: "wq : wquorums"
    shows "EX n . n : wq"
proof -
  def rq == "SOME rq . rq : rquorums"
  have f0: "rq : rquorums"
    apply (simp add: rq_def)
    apply (rule existenceSome)
    apply (insert rquorumsProp)
    apply (auto)
  done
  have f1: "rq ~= {}"
    apply (insert f0 rquorumsProp)
    apply (auto)
  done
  have f2: "ALL s'. s' : rquorums --> (EX n. n : wq & n : s')"
    apply (insert wquorums_def p0)
    apply (auto)
  done
  have f3: "EX n. n : wq & n : rq"
    by (simp add: f0 p0 quorumIntersect)
  show "?thesis"
    apply (insert f3)
    apply (blast)
  done
qed


theorem deadUnion1:
  "b : dead abstained ==> b :  dead (assign abstained i (abstained i Un B))"
  apply (simp add: dead_def)
  apply (auto)
done

theorem deadUnion:
  "dead abstained <= dead (assign abstained i (abstained i Un B))"
  apply (insert deadUnion1)
  apply (blast)
done


end

