(* Specification for automaton  Global1 *)

theory  Global1 = IOA2 + IOATools + Array + Set + AuxDist + Null:

datatype ModeType = idle | active

record Global1_state =
  initiated :: "(Node) Set"
  proposed :: "(Value) Set"
  decided :: "(Node) Set"
  failed :: "(Node) Set"
  ballots :: "(Ballot) Set"
  succeeded :: "(Ballot) Set"
  val :: "(Ballot, (Value) Null) Array"
  voted :: "(Node, (Ballot) Set) Array"
  abstained :: "(Node, (Ballot) Set) Array"

datatype Global1_action =
  init "Node" "Value"  |
  fail "Node"  |
  decide "Node" "Value" "Ballot"  |
  makeBallot "Ballot"  |
  abstain "Node" "(Ballot) Set"  |
  assignVal "Ballot" "Value"  |
  vote "Node" "Ballot"  |
  internalDecide "Ballot"  

consts
   Global1_start :: "Global1_state set"
   Global1_input :: "Global1_action set"
   Global1_output :: "Global1_action set"
   Global1_internal :: "Global1_action set"
   Global1_asig :: "Global1_action signature"
   Global1_effect :: "( Global1_action,  Global1_state) effect"
   Global1_enablement :: "( Global1_action,  Global1_state) enablement"
   Global1_trans :: "( Global1_action,  Global1_state) transition set"
   Global1 :: "( Global1_action,  Global1_state) ioa"

(* Warning: `so that' not implemented yet for initial state *)
defs
 Global1_start_def: 
  " Global1_start == { sGlobal1.
   (abstained  sGlobal1 = (constant {})) &
   (voted  sGlobal1 = (constant {})) &
   (val  sGlobal1 = (constant nil)) &
   (succeeded  sGlobal1 = {}) &
   (ballots  sGlobal1 = {}) &
   (failed  sGlobal1 = {}) &
   (decided  sGlobal1 = {}) &
   (proposed  sGlobal1 = {}) &
   (initiated  sGlobal1 = {})
  }"

(* Warning: `where' clauses in signature not implemented yet *)
defs
 Global1_input_def: 
  " Global1_input == {a. case a of 
    (init i v) => True |
    (fail i) => True |
    (decide i v b) => False |
    (makeBallot b) => False |
    (abstain i B) => False |
    (assignVal b v) => False |
    (vote i b) => False |
    (internalDecide b) => False
  }"
defs
 Global1_output_def: 
  " Global1_output == {a. case a of 
    (init i v) => False |
    (fail i) => False |
    (decide i v b) => True |
    (makeBallot b) => False |
    (abstain i B) => False |
    (assignVal b v) => False |
    (vote i b) => False |
    (internalDecide b) => False
  }"
defs
 Global1_internal_def: 
  " Global1_internal == {a. case a of 
    (init i v) => False |
    (fail i) => False |
    (decide i v b) => False |
    (makeBallot b) => True |
    (abstain i B) => True |
    (assignVal b v) => True |
    (vote i b) => True |
    (internalDecide b) => True
  }"
defs
   Global1_asig_def : " Global1_asig == ( Global1_input,  Global1_output,  Global1_internal)"

defs
 Global1_enablement_def: 
  " Global1_enablement  sGlobal1  aGlobal1 == case  aGlobal1 of
  (init i v) => True 
  
  | 
  (fail i) => True 
  
  | 
  (decide i v b) => True 
    & (((((i : (initiated sGlobal1)) & (~(i : (decided sGlobal1)))) & (~(i : (failed sGlobal1)))) & (b : (succeeded sGlobal1))) & ((embed v) = (sub (val sGlobal1) b)))
  | 
  (makeBallot b) => True 
    & (~(b : (ballots sGlobal1)))
  | 
  (abstain i B) => True 
    & (((i : (initiated sGlobal1)) & (~(i : (failed sGlobal1)))) & (((sub (voted sGlobal1) i) Int B) = {}))
  | 
  (assignVal b v) => True 
    & ((((b : (ballots sGlobal1)) & ((sub (val sGlobal1) b) = nil)) & (v : (proposed sGlobal1))) & (ALL b':: Ballot . ((b' < b) --> (((sub (val sGlobal1) b') = (embed v)) | (b' : (dead (abstained sGlobal1)))))))
  | 
  (vote i b) => True 
    & ((((i : (initiated sGlobal1)) & (~(i : (failed sGlobal1)))) & (b : (ballots sGlobal1))) & (~(b : (sub (abstained sGlobal1) i))))
  | 
  (internalDecide b) => True 
    & ((b : (ballots sGlobal1)) & (EX q_ID:: (Node) Set . (ALL j:: Node . ((q_ID : wquorums) & ((j : q_ID) --> (b : (sub (voted sGlobal1) j)))))))
  "
defs
 Global1_effect_def: 
  " Global1_effect  sGlobal1  aGlobal1 == case  aGlobal1 of
  (init i v) => 
    Global1_state.make 
     (if (~(i : (failed sGlobal1))) then (if (~(i : (initiated sGlobal1))) then ((initiated sGlobal1) Un {i}) else (initiated sGlobal1)) else (initiated sGlobal1))
     (if (~(i : (failed sGlobal1))) then (if (~(i : (initiated sGlobal1))) then ((proposed sGlobal1) Un {v}) else (proposed sGlobal1)) else (proposed sGlobal1))
     (decided sGlobal1)
     (failed sGlobal1)
     (ballots sGlobal1)
     (succeeded sGlobal1)
     (val sGlobal1)
     (voted sGlobal1)
     (abstained sGlobal1) | 
  (fail i) => 
    Global1_state.make 
     (initiated sGlobal1)
     (proposed sGlobal1)
     (decided sGlobal1)
     ((failed sGlobal1) Un {i})
     (ballots sGlobal1)
     (succeeded sGlobal1)
     (val sGlobal1)
     (voted sGlobal1)
     (abstained sGlobal1) | 
  (decide i v b) => 
    Global1_state.make 
     (initiated sGlobal1)
     (proposed sGlobal1)
     ((decided sGlobal1) Un {i})
     (failed sGlobal1)
     (ballots sGlobal1)
     (succeeded sGlobal1)
     (val sGlobal1)
     (voted sGlobal1)
     (abstained sGlobal1) | 
  (makeBallot b) => 
    Global1_state.make 
     (initiated sGlobal1)
     (proposed sGlobal1)
     (decided sGlobal1)
     (failed sGlobal1)
     ((ballots sGlobal1) Un {b})
     (succeeded sGlobal1)
     (val sGlobal1)
     (voted sGlobal1)
     (abstained sGlobal1) | 
  (abstain i B) => 
    Global1_state.make 
     (initiated sGlobal1)
     (proposed sGlobal1)
     (decided sGlobal1)
     (failed sGlobal1)
     (ballots sGlobal1)
     (succeeded sGlobal1)
     (val sGlobal1)
     (voted sGlobal1)
     (assign (abstained sGlobal1) i ((sub (abstained sGlobal1) i) Un B)) | 
  (assignVal b v) => 
    Global1_state.make 
     (initiated sGlobal1)
     (proposed sGlobal1)
     (decided sGlobal1)
     (failed sGlobal1)
     (ballots sGlobal1)
     (succeeded sGlobal1)
     (assign (val sGlobal1) b (embed v))
     (voted sGlobal1)
     (abstained sGlobal1) | 
  (vote i b) => 
    Global1_state.make 
     (initiated sGlobal1)
     (proposed sGlobal1)
     (decided sGlobal1)
     (failed sGlobal1)
     (ballots sGlobal1)
     (succeeded sGlobal1)
     (val sGlobal1)
     (assign (voted sGlobal1) i ((sub (voted sGlobal1) i) Un {b}))
     (abstained sGlobal1) | 
  (internalDecide b) => 
    Global1_state.make 
     (initiated sGlobal1)
     (proposed sGlobal1)
     (decided sGlobal1)
     (failed sGlobal1)
     (ballots sGlobal1)
     ((succeeded sGlobal1) Un {b})
     (val sGlobal1)
     (voted sGlobal1)
     (abstained sGlobal1)
  "
defs
 Global1_def: 
  " Global1 == (IOA  Global1_asig  Global1_start  Global1_enablement  Global1_effect)"

consts
Inv1 :: "Global1_state => bool"
Inv2 :: "Global1_state => bool"
Inv3 :: "Global1_state => bool"
Inv4 :: "Global1_state => bool"
Inv5 :: "Global1_state => bool"
Inv6 :: "Global1_state => bool"
Inv7 :: "Global1_state => bool"

defs
  Inv1_def : "Inv1 sGlobal1 == (ALL i:: Node . (ALL b:: Ballot . ((b : (sub (voted sGlobal1) i)) --> (~(b : (sub (abstained sGlobal1) i))))))"

defs
  Inv2_def : "Inv2 sGlobal1 == (ALL b:: Ballot . (((sub (val sGlobal1) b) ~= nil) --> ((Null.val (sub (val sGlobal1) b)) : (proposed sGlobal1))))"

defs
  Inv3_def : "Inv3 sGlobal1 == (ALL b:: Ballot . ((b : (succeeded sGlobal1)) --> (~(b : (dead (abstained sGlobal1))))))"

defs
  Inv4_def : "Inv4 sGlobal1 == (ALL b:: Ballot . (ALL b':: Ballot . ((((sub (val sGlobal1) b) ~= nil) & (b' < b)) --> (((sub (val sGlobal1) b') = (sub (val sGlobal1) b)) | (b' : (dead (abstained sGlobal1)))))))"

defs
  Inv5_def : "Inv5 sGlobal1 == (ALL b:: Ballot . ((b : (succeeded sGlobal1)) --> (b : (ballots sGlobal1))))"

defs
  Inv6_def : "Inv6 sGlobal1 == (ALL b_Inv6:: Ballot . ((b_Inv6 : (succeeded sGlobal1)) --> (EX q_Inv6:: (Node) Set . (ALL n_Inv6:: Node . ((q_Inv6 : wquorums) & ((n_Inv6 : q_Inv6) --> (b_Inv6 : (sub (voted sGlobal1) n_Inv6))))))))"

defs
  Inv7_def : "Inv7 sGlobal1 == (ALL b_Inv7:: Ballot . ((~(b_Inv7 : (ballots sGlobal1))) --> ((sub (val sGlobal1) b_Inv7) = nil)))"


(*  Inv1 : (ALL i:: Node . (ALL b:: Ballot . ((b : (sub (voted sGlobal1) i)) --> (~(b : (sub (abstained sGlobal1) i)))))) *)
theorem  Inv1_start:
  "invariant_start Global1 Inv1"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Global1_def  Global1_start_def  Inv1_def)
done

theorem  Inv1_trans_init:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((init i v))"
shows " Inv1 (effects_of  Global1 sGlobal1((init i v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv1_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv1_trans_fail:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((fail i))"
shows " Inv1 (effects_of  Global1 sGlobal1((fail i)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv1_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv1_trans_decide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((decide i v b))"
shows " Inv1 (effects_of  Global1 sGlobal1((decide i v b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv1_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv1_trans_makeBallot:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((makeBallot b))"
shows " Inv1 (effects_of  Global1 sGlobal1((makeBallot b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv1_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv1_trans_abstain:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((abstain i B))"
shows " Inv1 (effects_of  Global1 sGlobal1((abstain i B)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv1_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv1_trans_assignVal:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((assignVal b v))"
shows " Inv1 (effects_of  Global1 sGlobal1((assignVal b v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv1_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv1_trans_vote:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((vote i b))"
shows " Inv1 (effects_of  Global1 sGlobal1((vote i b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv1_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv1_trans_internalDecide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((internalDecide b))"
shows " Inv1 (effects_of  Global1 sGlobal1((internalDecide b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv1_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv1_trans1:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv1 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1 aGlobal1"
shows " Inv1 (effects_of  Global1 sGlobal1 aGlobal1)"
  apply (insert p0 p1 p2)
  apply (cases  aGlobal1)
  apply (simp add:  Inv1_trans_init)
  apply (simp add:  Inv1_trans_fail)
  apply (simp add:  Inv1_trans_decide)
  apply (simp add:  Inv1_trans_makeBallot)
  apply (simp add:  Inv1_trans_abstain)
  apply (simp add:  Inv1_trans_assignVal)
  apply (simp add:  Inv1_trans_vote)
  apply (simp add:  Inv1_trans_internalDecide)
done

theorem  Inv1_trans:
  "invariant_trans  Global1 Inv1"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  Inv1_trans1)
done

theorem  Inv1_invariant:
  "invariant  Global1 Inv1"
  apply (rule invariantI)
  apply (simp_all add:  Inv1_start  Inv1_trans)
done

theorem  Inv1_reachable:
  "reachable  Global1 sGlobal1 ==>  Inv1 sGlobal1"
  apply (insert  Inv1_invariant)
  apply (auto intro: invariantReachable)
done

(*  Inv2 : (ALL b:: Ballot . (((sub (val sGlobal1) b) ~= nil) --> ((val (sub (val sGlobal1) b)) : (proposed sGlobal1)))) *)
theorem  Inv2_start:
  "invariant_start Global1 Inv2"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Global1_def  Global1_start_def  Inv2_def)
done

theorem  Inv2_trans_init:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((init i v))"
shows " Inv2 (effects_of  Global1 sGlobal1((init i v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv2_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv2_trans_fail:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((fail i))"
shows " Inv2 (effects_of  Global1 sGlobal1((fail i)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv2_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv2_trans_decide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((decide i v b))"
shows " Inv2 (effects_of  Global1 sGlobal1((decide i v b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv2_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv2_trans_makeBallot:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((makeBallot b))"
shows " Inv2 (effects_of  Global1 sGlobal1((makeBallot b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv2_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv2_trans_abstain:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((abstain i B))"
shows " Inv2 (effects_of  Global1 sGlobal1((abstain i B)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv2_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv2_trans_assignVal:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((assignVal b v))"
shows " Inv2 (effects_of  Global1 sGlobal1((assignVal b v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv2_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv2_trans_vote:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((vote i b))"
shows " Inv2 (effects_of  Global1 sGlobal1((vote i b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv2_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv2_trans_internalDecide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((internalDecide b))"
shows " Inv2 (effects_of  Global1 sGlobal1((internalDecide b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv2_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv2_trans1:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv2 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1 aGlobal1"
shows " Inv2 (effects_of  Global1 sGlobal1 aGlobal1)"
  apply (insert p0 p1 p2)
  apply (cases  aGlobal1)
  apply (simp add:  Inv2_trans_init)
  apply (simp add:  Inv2_trans_fail)
  apply (simp add:  Inv2_trans_decide)
  apply (simp add:  Inv2_trans_makeBallot)
  apply (simp add:  Inv2_trans_abstain)
  apply (simp add:  Inv2_trans_assignVal)
  apply (simp add:  Inv2_trans_vote)
  apply (simp add:  Inv2_trans_internalDecide)
done

theorem  Inv2_trans:
  "invariant_trans  Global1 Inv2"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  Inv2_trans1)
done

theorem  Inv2_invariant:
  "invariant  Global1 Inv2"
  apply (rule invariantI)
  apply (simp_all add:  Inv2_start  Inv2_trans)
done

theorem  Inv2_reachable:
  "reachable  Global1 sGlobal1 ==>  Inv2 sGlobal1"
  apply (insert  Inv2_invariant)
  apply (auto intro: invariantReachable)
done

(*  Inv5 : (ALL b:: Ballot . ((b : (succeeded sGlobal1)) --> (b : (ballots sGlobal1)))) *)
theorem  Inv5_start:
  "invariant_start Global1 Inv5"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Global1_def  Global1_start_def  Inv5_def)
done

theorem  Inv5_trans_init:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((init i v))"
shows " Inv5 (effects_of  Global1 sGlobal1((init i v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv5_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv5_trans_fail:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((fail i))"
shows " Inv5 (effects_of  Global1 sGlobal1((fail i)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv5_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv5_trans_decide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((decide i v b))"
shows " Inv5 (effects_of  Global1 sGlobal1((decide i v b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv5_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv5_trans_makeBallot:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((makeBallot b))"
shows " Inv5 (effects_of  Global1 sGlobal1((makeBallot b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv5_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv5_trans_abstain:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((abstain i B))"
shows " Inv5 (effects_of  Global1 sGlobal1((abstain i B)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv5_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv5_trans_assignVal:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((assignVal b v))"
shows " Inv5 (effects_of  Global1 sGlobal1((assignVal b v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv5_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv5_trans_vote:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((vote i b))"
shows " Inv5 (effects_of  Global1 sGlobal1((vote i b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv5_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv5_trans_internalDecide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((internalDecide b))"
shows " Inv5 (effects_of  Global1 sGlobal1((internalDecide b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv5_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv5_trans1:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv5 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1 aGlobal1"
shows " Inv5 (effects_of  Global1 sGlobal1 aGlobal1)"
  apply (insert p0 p1 p2)
  apply (cases  aGlobal1)
  apply (simp add:  Inv5_trans_init)
  apply (simp add:  Inv5_trans_fail)
  apply (simp add:  Inv5_trans_decide)
  apply (simp add:  Inv5_trans_makeBallot)
  apply (simp add:  Inv5_trans_abstain)
  apply (simp add:  Inv5_trans_assignVal)
  apply (simp add:  Inv5_trans_vote)
  apply (simp add:  Inv5_trans_internalDecide)
done

theorem  Inv5_trans:
  "invariant_trans  Global1 Inv5"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  Inv5_trans1)
done

theorem  Inv5_invariant:
  "invariant  Global1 Inv5"
  apply (rule invariantI)
  apply (simp_all add:  Inv5_start  Inv5_trans)
done

theorem  Inv5_reachable:
  "reachable  Global1 sGlobal1 ==>  Inv5 sGlobal1"
  apply (insert  Inv5_invariant)
  apply (auto intro: invariantReachable)
done

(*  Inv4 : (ALL b:: Ballot . (ALL b':: Ballot . ((((sub (val sGlobal1) b) ~= nil) & (b' < b)) --> (((sub (val sGlobal1) b') = (sub (val sGlobal1) b)) | (b' : (dead (abstained sGlobal1))))))) *)
theorem  Inv4_start:
  "invariant_start Global1 Inv4"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Global1_def  Global1_start_def  Inv4_def)
done

theorem  Inv4_trans_init:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((init i v))"
shows " Inv4 (effects_of  Global1 sGlobal1((init i v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv4_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv4_trans_fail:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((fail i))"
shows " Inv4 (effects_of  Global1 sGlobal1((fail i)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv4_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv4_trans_decide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((decide i v b))"
shows " Inv4 (effects_of  Global1 sGlobal1((decide i v b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv4_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv4_trans_makeBallot:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((makeBallot b))"
shows " Inv4 (effects_of  Global1 sGlobal1((makeBallot b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv4_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv4_trans_abstain:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((abstain i B))"
shows " Inv4 (effects_of  Global1 sGlobal1((abstain i B)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv4_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
proof -
assume
  a0: "ALL b b'.
          Global1_state.val sGlobal1 b ~= nil & b' < b -->
          Global1_state.val sGlobal1 b' = Global1_state.val sGlobal1 b | b' : dead (abstained sGlobal1)"
  show "ALL b b'.
           Global1_state.val sGlobal1 b ~= nil & b' < b -->
           Global1_state.val sGlobal1 b' = Global1_state.val sGlobal1 b |
           b' : dead (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B))"
  apply (auto)
proof -
  fix b b'
  assume a1: "Global1_state.val sGlobal1 b ~= nil"
     and a2: "b' < b"
     and a3: "b' ~: dead (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B))"
  have f1: "     Global1_state.val sGlobal1 b' = Global1_state.val sGlobal1 b | b' : dead (abstained sGlobal1)"
    by (simp add: a1 a2 a3 a0)
  have f2: "b' ~: dead (abstained sGlobal1)"
    apply (insert a3 deadUnion1)
    apply (auto)
  done
  show "Global1_state.val sGlobal1 b' = Global1_state.val sGlobal1 b"
    apply (insert f2 f1)
    apply (auto)
  done
qed
qed


theorem  Inv4_trans_assignVal:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((assignVal b v))"
shows " Inv4 (effects_of  Global1 sGlobal1((assignVal b v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv4_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (rule allI)
proof -
  fix ba
  assume a0: "ALL b b'.
                Global1_state.val sGlobal1 b ~= nil & b' < b -->
                Global1_state.val sGlobal1 b' = Global1_state.val sGlobal1 b | b' : dead (abstained sGlobal1)"
     and a1: "             b : ballots sGlobal1 &
             Global1_state.val sGlobal1 b = nil &
             v : proposed sGlobal1 &
             (ALL b'. b' < b --> Global1_state.val sGlobal1 b' = embed v | b' : dead (abstained sGlobal1))"
  have a2: "(ALL b'. b' < b --> Global1_state.val sGlobal1 b' = embed v | b' : dead (abstained sGlobal1))"
    by (simp add: a1)
  show "      b ~= ba -->
              Global1_state.val sGlobal1 ba ~= nil & b < ba -->
              embed v = Global1_state.val sGlobal1 ba | b : dead (abstained sGlobal1)"
proof (rule impI)
  assume a3: "b ~= ba"
  show "Global1_state.val sGlobal1 ba ~= nil & b < ba --> embed v = Global1_state.val sGlobal1 ba | b : dead (abstained sGlobal1)"
proof -
  have f0: "~(Global1_state.val sGlobal1 ba ~= nil & b < ba & b ~: dead (abstained sGlobal1))"
    sorry
  show "?thesis"
    apply (auto)
    apply (insert f0)
    apply (auto)
  done
qed
qed
qed

theorem  Inv4_trans_vote:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((vote i b))"
shows " Inv4 (effects_of  Global1 sGlobal1((vote i b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv4_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv4_trans_internalDecide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((internalDecide b))"
shows " Inv4 (effects_of  Global1 sGlobal1((internalDecide b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv4_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv4_trans1:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv4 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1 aGlobal1"
shows " Inv4 (effects_of  Global1 sGlobal1 aGlobal1)"
  apply (insert p0 p1 p2)
  apply (cases  aGlobal1)
  apply (simp add:  Inv4_trans_init)
  apply (simp add:  Inv4_trans_fail)
  apply (simp add:  Inv4_trans_decide)
  apply (simp add:  Inv4_trans_makeBallot)
  apply (simp add:  Inv4_trans_abstain)
  apply (simp add:  Inv4_trans_assignVal)
  apply (simp add:  Inv4_trans_vote)
  apply (simp add:  Inv4_trans_internalDecide)
done

theorem  Inv4_trans:
  "invariant_trans  Global1 Inv4"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  Inv4_trans1)
done

theorem  Inv4_invariant:
  "invariant  Global1 Inv4"
  apply (rule invariantI)
  apply (simp_all add:  Inv4_start  Inv4_trans)
done

theorem  Inv4_reachable:
  "reachable  Global1 sGlobal1 ==>  Inv4 sGlobal1"
  apply (insert  Inv4_invariant)
  apply (auto intro: invariantReachable)
done


(*  Inv6 : (ALL b_Inv6:: Ballot . ((b_Inv6 : (succeeded sGlobal1)) --> (EX q_Inv6:: (Node) Set . (ALL n_Inv6:: Node . ((q_Inv6 : wquorums) & ((n_Inv6 : q_Inv6) --> (b_Inv6 : (sub (voted sGlobal1) n_Inv6)))))))) *)
theorem  Inv6_start:
  "invariant_start Global1 Inv6"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Global1_def  Global1_start_def  Inv6_def)
done

theorem  Inv6_trans_init:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((init i v))"
shows " Inv6 (effects_of  Global1 sGlobal1((init i v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv6_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv6_trans_fail:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((fail i))"
shows " Inv6 (effects_of  Global1 sGlobal1((fail i)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv6_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv6_trans_decide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((decide i v b))"
shows " Inv6 (effects_of  Global1 sGlobal1((decide i v b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv6_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv6_trans_makeBallot:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((makeBallot b))"
shows " Inv6 (effects_of  Global1 sGlobal1((makeBallot b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv6_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv6_trans_abstain:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((abstain i B))"
shows " Inv6 (effects_of  Global1 sGlobal1((abstain i B)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv6_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv6_trans_assignVal:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((assignVal b v))"
shows " Inv6 (effects_of  Global1 sGlobal1((assignVal b v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv6_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv6_trans_vote:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((vote i b))"
shows " Inv6 (effects_of  Global1 sGlobal1((vote i b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv6_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
proof -
  fix b_Inv6
  assume a0: "b_Inv6 : succeeded sGlobal1"
    and  a1: "ALL b_Inv6.
             b_Inv6 : succeeded sGlobal1 -->
             (EX q_Inv6. q_Inv6 : wquorums & (ALL n_Inv6. n_Inv6 : q_Inv6 --> b_Inv6 : voted sGlobal1 n_Inv6))"
  have f0: "(EX q_Inv6. q_Inv6 : wquorums & (ALL n_Inv6. n_Inv6 : q_Inv6 --> b_Inv6 : voted sGlobal1 n_Inv6))"
    by (simp add: a0 a1)
  show "EX q_Inv6.
              q_Inv6 : wquorums &
              (ALL n_Inv6.
                  (i = n_Inv6 --> n_Inv6 : q_Inv6 --> b_Inv6 = b | b_Inv6 : voted sGlobal1 n_Inv6) &
                  (i ~= n_Inv6 --> n_Inv6 : q_Inv6 --> b_Inv6 : voted sGlobal1 n_Inv6))"
    apply (insert f0)
    apply (auto)
  done
qed


theorem  Inv6_trans_internalDecide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((internalDecide b))"
shows " Inv6 (effects_of  Global1 sGlobal1((internalDecide b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv6_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv6_trans1:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv6 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1 aGlobal1"
shows " Inv6 (effects_of  Global1 sGlobal1 aGlobal1)"
  apply (insert p0 p1 p2)
  apply (cases  aGlobal1)
  apply (simp add:  Inv6_trans_init)
  apply (simp add:  Inv6_trans_fail)
  apply (simp add:  Inv6_trans_decide)
  apply (simp add:  Inv6_trans_makeBallot)
  apply (simp add:  Inv6_trans_abstain)
  apply (simp add:  Inv6_trans_assignVal)
  apply (simp add:  Inv6_trans_vote)
  apply (simp add:  Inv6_trans_internalDecide)
done

theorem  Inv6_trans:
  "invariant_trans  Global1 Inv6"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  Inv6_trans1)
done

theorem  Inv6_invariant:
  "invariant  Global1 Inv6"
  apply (rule invariantI)
  apply (simp_all add:  Inv6_start  Inv6_trans)
done

theorem  Inv6_reachable:
  "reachable  Global1 sGlobal1 ==>  Inv6 sGlobal1"
  apply (insert  Inv6_invariant)
  apply (auto intro: invariantReachable)
done



(*  Inv3 : (ALL b:: Ballot . ((b : (succeeded sGlobal1)) --> (~(b : (dead (abstained sGlobal1)))))) *)
theorem  Inv3_start:
  "invariant_start Global1 Inv3"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Global1_def  Global1_start_def  Inv3_def)
done

theorem  Inv3_trans_init:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((init i v))"
shows " Inv3 (effects_of  Global1 sGlobal1((init i v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv3_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv3_trans_fail:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((fail i))"
shows " Inv3 (effects_of  Global1 sGlobal1((fail i)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv3_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv3_trans_decide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((decide i v b))"
shows " Inv3 (effects_of  Global1 sGlobal1((decide i v b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv3_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv3_trans_makeBallot:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((makeBallot b))"
shows " Inv3 (effects_of  Global1 sGlobal1((makeBallot b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv3_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv3_trans_abstain:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((abstain i B))"
shows " Inv3 (effects_of  Global1 sGlobal1((abstain i B)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv3_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
proof -
  assume a0: "ALL b. b : succeeded sGlobal1 --> b ~: dead (abstained sGlobal1)"
  have f0: "Inv1 sGlobal1"
   and f1: "Inv6 sGlobal1"
    by (simp_all add: Inv1_reachable Inv6_reachable p0)
  show "ALL b. b : succeeded sGlobal1 --> b ~: dead (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B))"
proof (rule allI)
  fix b
  show "b : succeeded sGlobal1 --> b ~: dead (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B))"
proof (auto)
assume a1: "b : succeeded sGlobal1"
   and a2: "b : dead (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B))"
  have f2': "(EX wq. ALL n_wq. wq : wquorums & (n_wq : wq --> b : voted sGlobal1 [ n_wq ]))"
    apply (insert f1 Inv6_def a1)
    apply (auto)
  done
  have f2: "(EX wq. ((wq : wquorums) & (ALL n_wq . n_wq : wq --> b : voted sGlobal1 [ n_wq ])))"
    apply (insert f2')
    apply (auto)
  done
  have f3': "EX rq. rq : rquorums & (ALL n_rq :rq. b : (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B)) [ n_rq ])"
    apply (insert a2 dead_def)
    apply (auto)
  done
  have f3: "EX rq. rq : rquorums & (ALL n_rq . n_rq : rq --> b : (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B)) [ n_rq ])"
    apply (insert f3')
    apply (blast)
  done
  def wq == "(SOME wq. ((wq : wquorums) & (ALL n_wq . n_wq : wq --> b : voted sGlobal1 [ n_wq ])))"
    have f4: "((wq : wquorums) & (ALL n_wq . n_wq : wq --> b : voted sGlobal1 [ n_wq ]))"
      apply (simp add: wq_def)
      apply (rule existenceSome)
      apply (insert f2)
      apply (auto)
    done
  def rq == "SOME rq. rq : rquorums & (ALL n_rq . n_rq : rq --> b : (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B)) [ n_rq ])"
    have f5: "rq : rquorums & (ALL n_rq . n_rq : rq --> b : (assign (abstained sGlobal1) i (abstained sGlobal1 i Un B)) [ n_rq ])"
      sorry (* I am *not* going through this, it's blindingly obvious and Isabelle is too stupid *)
  have f6: "EX n . n : wq & n : rq"
    
  show "False"
    apply (insert dead_def)

done

theorem  Inv3_trans_assignVal:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((assignVal b v))"
shows " Inv3 (effects_of  Global1 sGlobal1((assignVal b v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv3_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv3_trans_vote:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((vote i b))"
shows " Inv3 (effects_of  Global1 sGlobal1((vote i b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv3_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv3_trans_internalDecide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((internalDecide b))"
shows " Inv3 (effects_of  Global1 sGlobal1((internalDecide b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv3_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv3_trans1:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv3 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1 aGlobal1"
shows " Inv3 (effects_of  Global1 sGlobal1 aGlobal1)"
  apply (insert p0 p1 p2)
  apply (cases  aGlobal1)
  apply (simp add:  Inv3_trans_init)
  apply (simp add:  Inv3_trans_fail)
  apply (simp add:  Inv3_trans_decide)
  apply (simp add:  Inv3_trans_makeBallot)
  apply (simp add:  Inv3_trans_abstain)
  apply (simp add:  Inv3_trans_assignVal)
  apply (simp add:  Inv3_trans_vote)
  apply (simp add:  Inv3_trans_internalDecide)
done

theorem  Inv3_trans:
  "invariant_trans  Global1 Inv3"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  Inv3_trans1)
done

theorem  Inv3_invariant:
  "invariant  Global1 Inv3"
  apply (rule invariantI)
  apply (simp_all add:  Inv3_start  Inv3_trans)
done

theorem  Inv3_reachable:
  "reachable  Global1 sGlobal1 ==>  Inv3 sGlobal1"
  apply (insert  Inv3_invariant)
  apply (auto intro: invariantReachable)
done


(*  Inv7 : (ALL b_Inv7:: Ballot . ((~(b_Inv7 : (ballots sGlobal1))) --> ((sub (val sGlobal1) b_Inv7) = nil))) *)
theorem  Inv7_start:
  "invariant_start Global1 Inv7"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Global1_def  Global1_start_def  Inv7_def)
done

theorem  Inv7_trans_init:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((init i v))"
shows " Inv7 (effects_of  Global1 sGlobal1((init i v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv7_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv7_trans_fail:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((fail i))"
shows " Inv7 (effects_of  Global1 sGlobal1((fail i)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv7_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv7_trans_decide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((decide i v b))"
shows " Inv7 (effects_of  Global1 sGlobal1((decide i v b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv7_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv7_trans_makeBallot:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((makeBallot b))"
shows " Inv7 (effects_of  Global1 sGlobal1((makeBallot b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv7_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv7_trans_abstain:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((abstain i B))"
shows " Inv7 (effects_of  Global1 sGlobal1((abstain i B)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv7_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv7_trans_assignVal:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((assignVal b v))"
shows " Inv7 (effects_of  Global1 sGlobal1((assignVal b v)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv7_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv7_trans_vote:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((vote i b))"
shows " Inv7 (effects_of  Global1 sGlobal1((vote i b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv7_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv7_trans_internalDecide:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1((internalDecide b))"
shows " Inv7 (effects_of  Global1 sGlobal1((internalDecide b)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  Inv7_def  Global1_def  Global1_enablement_def  Global1_effect_def  Global1_state.make_def)
  apply (auto ?)
done

theorem  Inv7_trans1:
assumes
  p0: "reachable  Global1 sGlobal1"
  and p1: " Inv7 sGlobal1"
  and p2: "enablement_of  Global1 sGlobal1 aGlobal1"
shows " Inv7 (effects_of  Global1 sGlobal1 aGlobal1)"
  apply (insert p0 p1 p2)
  apply (cases  aGlobal1)
  apply (simp add:  Inv7_trans_init)
  apply (simp add:  Inv7_trans_fail)
  apply (simp add:  Inv7_trans_decide)
  apply (simp add:  Inv7_trans_makeBallot)
  apply (simp add:  Inv7_trans_abstain)
  apply (simp add:  Inv7_trans_assignVal)
  apply (simp add:  Inv7_trans_vote)
  apply (simp add:  Inv7_trans_internalDecide)
done

theorem  Inv7_trans:
  "invariant_trans  Global1 Inv7"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  Inv7_trans1)
done

theorem  Inv7_invariant:
  "invariant  Global1 Inv7"
  apply (rule invariantI)
  apply (simp_all add:  Inv7_start  Inv7_trans)
done

theorem  Inv7_reachable:
  "reachable  Global1 sGlobal1 ==>  Inv7 sGlobal1"
  apply (insert  Inv7_invariant)
  apply (auto intro: invariantReachable)
done
end
