theory Peterson_proofs = Peterson:

(* InvA(s) <=> \A p ((s.pc[p] = trying1 \/ s.pc[p] = trying2 \/ s.pc[p] = critical0 \/ s.pc[p] = critical1) => s.flag[p]) *)
(* InvE(s) <=> \A p ((s.pc[p] = trying2) => (s.pc[s.turn] = trying2)) *)

(* crit(ctr) <=> (ctr = critical0 \/ ctr = critical1) *)
(* InvM(s) <=>  (\A p \A p' ((p ~= p') => ~(crit(s.pc[p]) /\ crit(s.pc[p'])))) *)


constdefs other :: "ProcType => ProcType"
  "other p == if p = p0 then p1 else p0"


lemma pOther: "other p ~= p"
  proof (induct p)
  qed (simp add: other_def)+


lemma pMust: "p = p0 | p = p1"
  apply (induct "p")
  apply (blast+)
done

lemma pType1: "p ~= p0 --> p = p1"
  by (induct_tac p) (blast+)

lemma pType2: "p ~= p1 --> p = p0"
  by (induct_tac p) (blast+)


lemma InvA_step_1:
  shows "[| ALL p. (sub (pc s) p = trying1 --> sub (flag s) p) &
                   (sub (pc s) p = trying2 --> sub (flag s) p) &
                   (sub (pc s) p = critical0 --> sub (flag s) p) & (sub (pc s) p = critical1 --> sub (flag s) p);
            sub (pc s) p0 = trying2; turn s = p0; turn s' = p0; flag s' = flag s; ~ sub (flag s) p1;
            pc s' = assign (pc s) p0 critical0; p0 ~= p; sub (pc s) p = trying2 |]
         ==> False"
    proof -
    assume
        a0:"sub (pc s) p0 = trying2"
    and a1:"turn s = p0 "
    and a2:"turn s' = p0 "
    and a3:"flag s' = flag s " 
    and a4:"~ sub (flag s) p1 " 
    and a5:"pc s' = assign (pc s) p0 critical0 " 
    and a6:"p0 ~= p"
    and a7:"sub (pc s) p = trying2"
    have step2: "p ~= p0"
      apply (insert a6)
      apply (blast)
    done
    have step3: "p = p1"
      apply (insert pType1 step2)
      apply (blast)
    done
    have step4: "sub (pc s) p1 = trying2"
      apply (insert step3 a7)
      apply (blast)
    done
    note step4 and a4
    thus "ALL p. (sub (pc s) p = trying1 --> sub (flag s) p) &
                   (sub (pc s) p = trying2 --> sub (flag s) p) &
                   (sub (pc s) p = critical0 --> sub (flag s) p) & (sub (pc s) p = critical1 --> sub (flag s) p)
          ==> False"
      by (auto)
    qed

lemma InvA_step_2:
  shows "[| ALL p. (sub (pc s) p = trying1 --> sub (flag s) p) &
                   (sub (pc s) p = trying2 --> sub (flag s) p) &
                   (sub (pc s) p = critical0 --> sub (flag s) p) & (sub (pc s) p = critical1 --> sub (flag s) p);
            sub (pc s) (turn s) = trying2; turn s' = turn s; flag s' = flag s; turn s ~= p0; ~ sub (flag s) p0;
            pc s' = assign (pc s) (turn s) critical0; turn s ~= p; sub (pc s) p = trying2 |]
         ==> False"
      proof -

      assume 
	    a0: "sub (pc s) (turn s) = trying2"
	and a1: "turn s' = turn s"
	and a2: "flag s' = flag s"
	and a3: "turn s ~= p0"
	and a4: "~ sub (flag s) p0"
	and a5: "pc s' = assign (pc s) (turn s) critical0"
	and a6: "turn s ~= p"
	and a7: "sub (pc s) p = trying2"
    have step2: "turn s = p1"
      apply (insert a3 pType2)
      apply (blast)
    done
    have step3: "p1 ~= p"
      apply (insert step2 a6)
      apply (blast)
    done
    have step4: "p = p0"
      apply (insert step3 pType1)
      apply (blast)
    done
    have step4: "sub (pc s) p0 = trying2"
      apply (insert step4 a7)
      apply (blast)
    done
    note step4 and a4

    thus "[| ALL p. (sub (pc s) p = trying1 --> sub (flag s) p) &
      (sub (pc s) p = trying2 --> sub (flag s) p) &
      (sub (pc s) p = critical0 --> sub (flag s) p) & (sub (pc s) p = critical1 --> sub (flag s) p)|]
      ==> False"
      by (auto)
    qed



constdefs InvA :: "Peterson_state => bool"
  "InvA s == ALL p::ProcType . 
    (  ((sub (pc s) p) = trying1) |
       ((sub (pc s) p) = trying2) | 
       ((sub (pc s) p) = critical0) | 
       ((sub (pc s) p) = critical1)
    ) --> 
       (sub (flag s) p)
  "

constdefs InvE :: "Peterson_state => bool"
  "InvE s ==  (EX p. sub (pc s) p = trying2) --> sub (pc s) (turn s) = trying2
  "
(* ALL p::ProcType .     (sub (pc s) p) = trying2 --> (sub (pc s) (turn s)) = trying2 *)


constdefs InvM :: "Peterson_state => bool"
  "InvM s == ALL p::ProcType . 
    ((sub (pc s) p) = critical0 --> (sub (pc s) (other p)) ~= critical0 & (sub (pc s) (other p)) ~= critical1) &
    ((sub (pc s) p) = critical1 --> (sub (pc s) (other p)) ~= critical0 & (sub (pc s) (other p)) ~= critical1) 

  "

theorem InvA_start:
 "s : Peterson_start ==> InvA s"
  apply (simp add: Peterson_start_def) 
  proof -
  assume "turn s = (if randomBool then p0 else p1) & flag s = constant False & pc s = constant waiting0"
  show "InvA s"
    proof (simp add: InvA_def prems) qed
qed

theorem InvE_start:
 "s : Peterson_start ==> InvE s"
  apply (simp add: Peterson_start_def) 
  apply (auto)
  apply (simp add: InvE_def)
done


(* This is the other way of doing it.  I don't like it as much because
it's not organized *)

theorem InvM_start:
 "s : Peterson_start ==> InvM s"
  apply (simp add: Peterson_start_def) 
  proof -
  assume "turn s = (if randomBool then p0 else p1) & flag s = constant False & pc s = constant waiting0"
  show "InvM s"
    by (simp add: InvM_def prems) 
qed

theorem InvA_trans_trying:
 "InvA s & (s, trying p, s') : Peterson_trans ==> InvA s'"
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvA_def)
done

theorem InvA_trans:
 "InvA s & (s, act, s') : Peterson_trans --> InvA s'"
  proof (cases "act")
  {case trying
   show "?thesis"
   apply (simp_all add: Peterson_trans_def prems)
   apply (simp_all add: InvA_def)
   done
  }
  qed (simp_all add: Peterson_trans_def InvA_def)
(* Shown using the cases-induction method *)


theorem InvA_trans:
 "InvA s & (s, act, s') : Peterson_trans --> InvA s'"
  apply (induct_tac act)
  apply (simp_all add: Peterson_trans_def)
  apply (simp_all add: InvA_def)
done (* Holy SHIT *)


theorem InvE_trans:
 "InvE s & InvA s & (s, act, s') : Peterson_trans --> InvE s'"
  apply (induct_tac act)

  (* trying *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvE_def InvA_def)
  apply (auto)

  (* setFlag *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvE_def)
  apply (auto)

  (* setTurn *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvE_def)
  apply (auto?)

  (* checkFlag *)
  defer

  (* checkTurn *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvE_def)
  apply (auto?)

  (* critical *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvE_def)
  apply (auto?)

  (* release *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvE_def)
  apply (auto?)

  (* checkFlag *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvA_def)
  apply (simp add: InvE_def)
  apply (auto?)

  apply (auto intro: InvA_step_1)
  apply (auto intro: InvA_step_2)
done
(* Ugh ugh ugh! *)

theorem InvM_trans_checkFlag_1:
 "[| ALL p. (sub (pc s) p = trying1 --> sub (flag s) p) &
                           (sub (pc s) p = trying2 --> sub (flag s) p) & (sub (pc s) p = critical0 --> sub (flag s) p) & (sub (pc s) p = critical1 --> sub (flag s) p);
            sub (pc s) p0 = trying2; turn s' = turn s; flag s' = flag s; ~ sub (flag s) p1; pc s' = assign (pc s) p0 critical0; p ~= p0; p0 ~= p;
            sub (pc s) p = critical0 |]
         ==> False"
  proof -

  assume  a0: "sub (pc s) p0 = trying2"
    and a1: "flag s' = flag s"
    and a2: "~ sub (flag s) p1"
    and a3: "pc s' = assign (pc s) p0 critical0"
    and a4: "p ~= p0"

  have step3: "p = p1"
    apply (insert pType1 a4)
    apply (blast)
    done

  assume a5: "sub (pc s) p = critical0"
     and a6: "ALL p. (sub (pc s) p = trying1 --> sub (flag s) p) &
      (sub (pc s) p = trying2 --> sub (flag s) p) & (sub (pc s) p = critical0 --> sub (flag s) p) & (sub (pc s) p = critical1 --> sub (flag s) p)"

  have step4: "sub (pc s) p1 = critical0"
    apply (insert step3 a5)
    apply (blast)
  done
  note step4 and a2 and a6
  thus "False"
    by (auto)
  qed

theorem InvM_trans_checkFlag_2:
 "[| ALL p. (sub (pc s) p = trying1 --> sub (flag s) p) &
                           (sub (pc s) p = trying2 --> sub (flag s) p) & (sub (pc s) p = critical0 --> sub (flag s) p) & (sub (pc s) p = critical1 --> sub (flag s) p);
            sub (pc s) p0 = trying2; turn s' = turn s; flag s' = flag s; ~ sub (flag s) p1; pc s' = assign (pc s) p0 critical0; p ~= p0; p0 ~= p;
            sub (pc s) p = critical1 |]
         ==> False"
  proof -

  assume  a0: "sub (pc s) p0 = trying2"
    and a1: "flag s' = flag s"
    and a2: "~ sub (flag s) p1"
    and a3: "pc s' = assign (pc s) p0 critical0"
    and a4: "p ~= p0"

  have step3: "p = p1"
    apply (insert pType1 a4)
    apply (blast)
    done

  assume a5: "sub (pc s) p = critical1"
     and a6: "ALL p. (sub (pc s) p = trying1 --> sub (flag s) p) &
      (sub (pc s) p = trying2 --> sub (flag s) p) & (sub (pc s) p = critical0 --> sub (flag s) p) & (sub (pc s) p = critical1 --> sub (flag s) p)"

  have step4: "sub (pc s) p1 = critical1"
    apply (insert step3 a5)
    apply (blast)
  done
  note step4 and a2 and a6
  thus "False"
    by (auto)
  qed


theorem InvM_trans_checkFlag:
 "[| InvM s; InvE s; InvA s; (s, checkFlag pInv, s') : Peterson_trans |] ==> InvM s'"
  apply (simp_all add: Peterson_trans_def)
  apply (simp_all add: InvM_def InvA_def)
  apply (simp_all add: other_def)
  apply (auto)
  (* Auto did the case analysis for p0, p1 *)

  apply (auto intro: InvM_trans_checkFlag_1)
  apply (auto intro: InvM_trans_checkFlag_2)
done

theorem InvM_trans_checkTurn:
  "[| InvM s; InvE s; InvA s; (s, checkTurn pInv, s') : Peterson_trans |] ==> InvM s'"
  proof -

  assume "InvM s"
  have InvM_s: 
"ALL p. (sub (pc s) p = critical0 --> sub (pc s) (other p) ~= critical0 & sub (pc s) (other p) ~= critical1) &
        (sub (pc s) p = critical1 --> sub (pc s) (other p) ~= critical0 & sub (pc s) (other p) ~= critical1)"
    apply (insert InvM_def prems)
    apply (blast) (* Why not simp?? *)
  done
    have f00: "(sub (pc s) p0 = critical0 --> sub (pc s) (other p0) ~= critical0 & sub (pc s) (other p0) ~= critical1)"
      by (simp add: InvM_s)
    have f01: "(sub (pc s) p0 = critical1 --> sub (pc s) (other p0) ~= critical0 & sub (pc s) (other p0) ~= critical1)"
      by (simp add: InvM_s)
    have f10: "(sub (pc s) p1 = critical0 --> sub (pc s) (other p1) ~= critical0 & sub (pc s) (other p1) ~= critical1)"
      by (simp add: InvM_s)
    have f11: "(sub (pc s) p1 = critical1 --> sub (pc s) (other p1) ~= critical0 & sub (pc s) (other p1) ~= critical1)"
      by (simp add: InvM_s)

  assume a0: "InvE s"
     and a1: "InvA s"
     and a2: "(s, checkTurn pInv, s') : Peterson_trans"
  have a3: "sub (pc s) pInv = trying2"
   and a4: "turn s' = turn s"
   and a5: "flag s' = flag s"
   and a6: "pc s' = (if turn s ~= pInv then assign (pc (fst (s, checkTurn pInv, s'))) pInv critical0 else pc (fst (s, checkTurn pInv, s')))"
    apply (insert a2)
    apply (simp_all add: Peterson_trans_def)
  done
  show "InvM s'"
  proof (cases "turn s = pInv")
  {case True
   have a7: "pc s' = pc s"
     by (simp add: prems a6)
   show "InvM s'"
     apply (simp add: InvM_def)
     apply (insert InvM_s a7)
     apply (auto)
   done
  }

  {case False
   
   have a7: "pc s' = assign (pc s) pInv critical0"
     by (simp add: prems a6)
   note a0 and InvE_def and a3
   hence a8: "sub (pc s) (turn s) = trying2"
     by (blast)
   have a9: "turn s = other pInv"
     proof (cases pInv)
     {case p0
      have a10: "turn s = p1"
	apply (insert False pType1 p0)
	apply (blast)
      done
      show "?thesis"
	apply (simp add: other_def a10 p0)
      done
     }
     {case p1
      have a10: "turn s = p0"
	apply (insert False pType1 p1)
	apply (blast)
      done
      show "?thesis"
	apply (simp add: other_def a10 p1)
      done
     }
   qed
   have a10: "sub (pc s) (other pInv) = trying2"
     apply (insert a9 a8)
     apply (simp)
   done
   have a11: "sub (pc s') pInv = critical0"
     apply (simp add: a7 sub_def)
   done
   have a12: "sub (pc s') (other pInv) = trying2"
     apply (simp add: a7 sub_def a10 )
     apply (simp add: pOther other_def)
   done

   show "InvM s'"
   apply (simp add: InvM_def)
   apply (rule allI)   
   proof -
   fix p
   show "(sub (pc s') p = critical0 --> sub (pc s') (other p) ~= critical0 & sub (pc s') (other p) ~= critical1) &
         (sub (pc s') p = critical1 --> sub (pc s') (other p) ~= critical0 & sub (pc s') (other p) ~= critical1)"
   proof (cases "p = pInv")
   {case True
    show "?thesis"
      apply (auto)
      apply (simp_all add: True a11)
      apply (simp_all add: a12)
    done }
   {case False
    have a13: "p = (other pInv)"
     proof (cases pInv)
     {case p0
      have a14: "(other pInv) = p1"
	apply (simp add: False pType1 p0 other_def)
      done
      show "?thesis"
	apply (insert p0 a14 False)
	apply (simp add: pType1)
      done
     }
     {case p1
      have a14: "(other pInv) = p0"
	apply (simp add: False pType1 p1 other_def)
      done
      show "?thesis"
	apply (insert p1 a14 False)
	apply (simp add: pType2)
      done
     }
     qed

    show "?thesis"
      apply (auto)
      apply (simp_all add: False a13)
      apply (simp_all add: a12)
    done
   }
   qed
  qed }
qed qed


theorem InvM_trans:
 "InvM s & InvE s & InvA s & (s, act, s') : Peterson_trans --> InvM s'"
  apply (induct_tac act)
  apply (auto)


  (* trying *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvM_def)
  apply (auto?)

  (* setFlag *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvM_def)
  apply (auto?)

  (* setTurn *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvM_def)
  apply (auto?)

  (* checkFlag *)
  apply (auto intro: InvM_trans_checkFlag)

  (* checkTurn *)
  apply (auto intro: InvM_trans_checkTurn)

  (* critical *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvM_def)
  apply (auto?)

  (* release *)
  apply (simp add: Peterson_trans_def)
  apply (simp add: InvM_def)
  apply (auto?)

done


(* Ugh, why can't I put step 3 and 4 after the "show" and still have it work?? *)
theorem InvA: "invariant Peterson InvA"
  proof -
  have step1: "starts_of Peterson = Peterson_start"
    by (simp add: Peterson_def starts_of_def)

  have step2: "trans_of Peterson = Peterson_trans"
    by (simp add: Peterson_def trans_of_def)

  have step3: "!!s. s : starts_of Peterson ==> InvA s"
    apply (simp add: step1 InvA_start)
  done

  have step4: "!!s s' act. InvA(s) ==> (s,act,s'): trans_of(Peterson) --> InvA(s')"
    apply (insert step2 InvA_trans)
    apply (blast)
  done

  show "?thesis"
    apply (rule invariantI)
    apply (simp add: step3)
    apply (insert step4)
    apply (blast)
  done
qed

(* Why define AEM?  Cuz M by itself doesn't go through *)

constdefs InvAEM :: "Peterson_state => bool"
  "InvAEM s == InvA s & InvE s & InvM s"

lemma InvAEM_start:
 "s : Peterson_start ==> InvAEM s"
  apply (simp add: InvAEM_def)
  apply (auto)
  apply (simp_all add: InvA_start InvE_start InvM_start)
done

lemma InvAEM_trans:
 "InvAEM s & (s, act, s') : Peterson_trans ==> InvAEM s'"
  apply (simp add: InvAEM_def)
  apply (auto)
  apply (insert InvA_trans InvE_trans InvM_trans)
  apply (blast+)
done




(* Ugh, why can't I put step 3 and 4 after the "show" and still have it work?? *)
theorem InvAEM: "invariant Peterson InvAEM"
  proof -
  have step1: "starts_of Peterson = Peterson_start"
    by (simp add: Peterson_def starts_of_def)

  have step2: "trans_of Peterson = Peterson_trans"
    by (simp add: Peterson_def trans_of_def)

  have step3: "!!s. s : starts_of Peterson ==> InvAEM s"
    apply (simp add: step1 InvAEM_start)
  done

  have step4: "!!s s' act. InvAEM(s) ==> (s,act,s'): trans_of(Peterson) --> InvAEM(s')"
    apply (insert step2 InvAEM_trans)
    apply (blast)
  done

  show "?thesis"
    apply (rule invariantI)
    apply (simp add: step3)
    apply (insert step4)
    apply (blast)
  done
qed


