(* Specification for automaton  Peterson *)

theory  Peterson = IOA2 + IOATools + Array + NonDet:

datatype ProcType = p0 | p1
datatype PCType = waiting0 | trying0 | trying1 | trying2 | critical0 | critical1

record Peterson_state =
  pc :: "(ProcType, PCType) Array"
  flag :: "(ProcType, bool) Array"
  turn :: "ProcType"

datatype Peterson_action =
  trying "ProcType"  |
  setFlag "ProcType"  |
  setTurn "ProcType"  |
  checkFlag "ProcType"  |
  checkTurn "ProcType"  |
  critical "ProcType"  |
  release "ProcType"  

consts
   Peterson_start :: "Peterson_state set"
   Peterson_input :: "Peterson_action set"
   Peterson_output :: "Peterson_action set"
   Peterson_internal :: "Peterson_action set"
   Peterson_asig :: "Peterson_action signature"
   Peterson_effect :: "( Peterson_action,  Peterson_state) effect"
   Peterson_enablement :: "( Peterson_action,  Peterson_state) enablement"
   Peterson_trans :: "( Peterson_action,  Peterson_state) transition set"
   Peterson :: "( Peterson_action,  Peterson_state) ioa"

(* Warning: `so that' not implemented yet for initial state *)
defs
 Peterson_start_def: 
  " Peterson_start == { sPeterson.
   (turn  sPeterson = (if randomBool then p0 else p1)) &
   (flag  sPeterson = (constant false)) &
   (pc  sPeterson = (constant waiting0))
  }"

(* Warning: `where' clauses in signature not implemented yet *)
defs
 Peterson_input_def: 
  " Peterson_input == {a. case a of 
    (trying p) => False |
    (setFlag p) => False |
    (setTurn p) => False |
    (checkFlag p) => False |
    (checkTurn p) => False |
    (critical p) => False |
    (release p) => False
  }"
defs
 Peterson_output_def: 
  " Peterson_output == {a. case a of 
    (trying p) => True |
    (setFlag p) => False |
    (setTurn p) => False |
    (checkFlag p) => False |
    (checkTurn p) => False |
    (critical p) => True |
    (release p) => True
  }"
defs
 Peterson_internal_def: 
  " Peterson_internal == {a. case a of 
    (trying p) => False |
    (setFlag p) => True |
    (setTurn p) => True |
    (checkFlag p) => True |
    (checkTurn p) => True |
    (critical p) => False |
    (release p) => False
  }"
defs
   Peterson_asig_def : " Peterson_asig == ( Peterson_input,  Peterson_output,  Peterson_internal)"

defs
 Peterson_enablement_def: 
  " Peterson_enablement  sPeterson  aPeterson == case  aPeterson of
  (trying p) => True 
    & ((sub (pc sPeterson) p) = waiting0)
  | 
  (setFlag p) => True 
    & ((sub (pc sPeterson) p) = trying0)
  | 
  (setTurn p) => True 
    & ((sub (pc sPeterson) p) = trying1)
  | 
  (checkFlag p) => True 
    & ((sub (pc sPeterson) p) = trying2)
  | 
  (checkTurn p) => True 
    & ((sub (pc sPeterson) p) = trying2)
  | 
  (critical p) => True 
    & ((sub (pc sPeterson) p) = critical0)
  | 
  (release p) => True 
    & ((sub (pc sPeterson) p) = critical1)
  "
defs
 Peterson_effect_def: 
  " Peterson_effect  sPeterson  aPeterson == case  aPeterson of
  (trying p) => 
    Peterson_state.make 
     (assign (pc sPeterson) p trying0)
     (flag sPeterson)
     (turn sPeterson) | 
  (setFlag p) => 
    Peterson_state.make 
     (assign (pc sPeterson) p trying1)
     (assign (flag sPeterson) p true)
     (turn sPeterson) | 
  (setTurn p) => 
    Peterson_state.make 
     (assign (pc sPeterson) p trying2)
     (flag sPeterson)
     p | 
  (checkFlag p) => 
    Peterson_state.make 
     (if ((sub (flag sPeterson) (if (p = p0) then p1 else p0)) = false) then (assign (pc sPeterson) p critical0) else (pc sPeterson))
     (flag sPeterson)
     (turn sPeterson) | 
  (checkTurn p) => 
    Peterson_state.make 
     (if ((turn sPeterson) ~= p) then (assign (pc sPeterson) p critical0) else (pc sPeterson))
     (flag sPeterson)
     (turn sPeterson) | 
  (critical p) => 
    Peterson_state.make 
     (assign (pc sPeterson) p critical1)
     (flag sPeterson)
     (turn sPeterson) | 
  (release p) => 
    Peterson_state.make 
     (assign (pc sPeterson) p waiting0)
     (assign (flag sPeterson) p false)
     (turn sPeterson)
  "
defs
 Peterson_def: 
  " Peterson == (IOA  Peterson_asig  Peterson_start  Peterson_enablement  Peterson_effect)"

consts
InvA :: "Peterson_state => bool"
InvE :: "Peterson_state => bool"
InvM :: "Peterson_state => bool"

defs
  InvA_def : "InvA sPeterson == (ALL p:: ProcType . ((((((sub (pc sPeterson) p) = trying1) | ((sub (pc sPeterson) p) = trying2)) | ((sub (pc sPeterson) p) = critical0)) | ((sub (pc sPeterson) p) = critical1)) --> (sub (flag sPeterson) p)))"

defs
  InvE_def : "InvE sPeterson == ((EX p:: ProcType . ((sub (pc sPeterson) p) = trying2)) --> ((sub (pc sPeterson) (turn sPeterson)) = trying2))"

defs
  InvM_def : "InvM sPeterson == (ALL p1:: ProcType . (ALL p2:: ProcType . ((p1 ~= p2) --> ((((sub (pc sPeterson) p1) = critical0) --> (((sub (pc sPeterson) p2) ~= critical0) & ((sub (pc sPeterson) p2) ~= critical1))) & (((sub (pc sPeterson) p1) = critical1) --> (((sub (pc sPeterson) p2) ~= critical0) & ((sub (pc sPeterson) p2) ~= critical1)))))))"


(*  InvA : (ALL p:: ProcType . ((((((sub (pc sPeterson) p) = trying1) | ((sub (pc sPeterson) p) = trying2)) | ((sub (pc sPeterson) p) = critical0)) | ((sub (pc sPeterson) p) = critical1)) --> (sub (flag sPeterson) p))) *)
theorem  InvA_start:
  "invariant_start Peterson InvA"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Peterson_def  Peterson_start_def  InvA_def)
done

theorem  InvA_trans_trying:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvA sPeterson"
  and p2: "enablement_of  Peterson sPeterson((trying p))"
shows " InvA (effects_of  Peterson sPeterson((trying p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvA_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvA_trans_setFlag:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvA sPeterson"
  and p2: "enablement_of  Peterson sPeterson((setFlag p))"
shows " InvA (effects_of  Peterson sPeterson((setFlag p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvA_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvA_trans_setTurn:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvA sPeterson"
  and p2: "enablement_of  Peterson sPeterson((setTurn p))"
shows " InvA (effects_of  Peterson sPeterson((setTurn p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvA_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvA_trans_checkFlag:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvA sPeterson"
  and p2: "enablement_of  Peterson sPeterson((checkFlag p))"
shows " InvA (effects_of  Peterson sPeterson((checkFlag p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvA_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvA_trans_checkTurn:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvA sPeterson"
  and p2: "enablement_of  Peterson sPeterson((checkTurn p))"
shows " InvA (effects_of  Peterson sPeterson((checkTurn p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvA_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvA_trans_critical:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvA sPeterson"
  and p2: "enablement_of  Peterson sPeterson((critical p))"
shows " InvA (effects_of  Peterson sPeterson((critical p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvA_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvA_trans_release:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvA sPeterson"
  and p2: "enablement_of  Peterson sPeterson((release p))"
shows " InvA (effects_of  Peterson sPeterson((release p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvA_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvA_trans1:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvA sPeterson"
  and p2: "enablement_of  Peterson sPeterson aPeterson"
shows " InvA (effects_of  Peterson sPeterson aPeterson)"
  apply (insert p0 p1 p2)
  apply (cases  aPeterson)
  apply (simp add:  InvA_trans_trying)
  apply (simp add:  InvA_trans_setFlag)
  apply (simp add:  InvA_trans_setTurn)
  apply (simp add:  InvA_trans_checkFlag)
  apply (simp add:  InvA_trans_checkTurn)
  apply (simp add:  InvA_trans_critical)
  apply (simp add:  InvA_trans_release)
done

theorem  InvA_trans:
  "invariant_trans  Peterson InvA"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  InvA_trans1)
done

theorem  InvA_invariant:
  "invariant  Peterson InvA"
  apply (rule invariantI)
  apply (simp_all add:  InvA_start  InvA_trans)
done

theorem  InvA_reachable:
  "reachable  Peterson sPeterson ==>  InvA sPeterson"
  apply (insert  InvA_invariant)
  apply (auto intro: invariantReachable)
done

(*  InvE : ((EX p:: ProcType . ((sub (pc sPeterson) p) = trying2)) --> ((sub (pc sPeterson) (turn sPeterson)) = trying2)) *)
theorem  InvE_start:
  "invariant_start Peterson InvE"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Peterson_def  Peterson_start_def  InvE_def)
done

theorem  InvE_trans_trying:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvE sPeterson"
  and p2: "enablement_of  Peterson sPeterson((trying p))"
shows " InvE (effects_of  Peterson sPeterson((trying p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvE_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvE_trans_setFlag:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvE sPeterson"
  and p2: "enablement_of  Peterson sPeterson((setFlag p))"
shows " InvE (effects_of  Peterson sPeterson((setFlag p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvE_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvE_trans_setTurn:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvE sPeterson"
  and p2: "enablement_of  Peterson sPeterson((setTurn p))"
shows " InvE (effects_of  Peterson sPeterson((setTurn p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvE_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvE_trans_checkFlag:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvE sPeterson"
  and p2: "enablement_of  Peterson sPeterson((checkFlag p))"
shows " InvE (effects_of  Peterson sPeterson((checkFlag p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvE_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvE_trans_checkTurn:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvE sPeterson"
  and p2: "enablement_of  Peterson sPeterson((checkTurn p))"
shows " InvE (effects_of  Peterson sPeterson((checkTurn p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvE_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvE_trans_critical:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvE sPeterson"
  and p2: "enablement_of  Peterson sPeterson((critical p))"
shows " InvE (effects_of  Peterson sPeterson((critical p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvE_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvE_trans_release:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvE sPeterson"
  and p2: "enablement_of  Peterson sPeterson((release p))"
shows " InvE (effects_of  Peterson sPeterson((release p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvE_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvE_trans1:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvE sPeterson"
  and p2: "enablement_of  Peterson sPeterson aPeterson"
shows " InvE (effects_of  Peterson sPeterson aPeterson)"
  apply (insert p0 p1 p2)
  apply (cases  aPeterson)
  apply (simp add:  InvE_trans_trying)
  apply (simp add:  InvE_trans_setFlag)
  apply (simp add:  InvE_trans_setTurn)
  apply (simp add:  InvE_trans_checkFlag)
  apply (simp add:  InvE_trans_checkTurn)
  apply (simp add:  InvE_trans_critical)
  apply (simp add:  InvE_trans_release)
done

theorem  InvE_trans:
  "invariant_trans  Peterson InvE"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  InvE_trans1)
done

theorem  InvE_invariant:
  "invariant  Peterson InvE"
  apply (rule invariantI)
  apply (simp_all add:  InvE_start  InvE_trans)
done

theorem  InvE_reachable:
  "reachable  Peterson sPeterson ==>  InvE sPeterson"
  apply (insert  InvE_invariant)
  apply (auto intro: invariantReachable)
done

(*  InvM : (ALL p1:: ProcType . (ALL p2:: ProcType . ((p1 ~= p2) --> ((((sub (pc sPeterson) p1) = critical0) --> (((sub (pc sPeterson) p2) ~= critical0) & ((sub (pc sPeterson) p2) ~= critical1))) & (((sub (pc sPeterson) p1) = critical1) --> (((sub (pc sPeterson) p2) ~= critical0) & ((sub (pc sPeterson) p2) ~= critical1))))))) *)
theorem  InvM_start:
  "invariant_start Peterson InvM"
  apply (simp add: invariant_start_def)
  apply (auto )
  apply (simp add:  Peterson_def  Peterson_start_def  InvM_def)
done

theorem  InvM_trans_trying:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvM sPeterson"
  and p2: "enablement_of  Peterson sPeterson((trying p))"
shows " InvM (effects_of  Peterson sPeterson((trying p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvM_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvM_trans_setFlag:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvM sPeterson"
  and p2: "enablement_of  Peterson sPeterson((setFlag p))"
shows " InvM (effects_of  Peterson sPeterson((setFlag p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvM_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvM_trans_setTurn:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvM sPeterson"
  and p2: "enablement_of  Peterson sPeterson((setTurn p))"
shows " InvM (effects_of  Peterson sPeterson((setTurn p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvM_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvM_trans_checkFlag:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvM sPeterson"
  and p2: "enablement_of  Peterson sPeterson((checkFlag p))"
shows " InvM (effects_of  Peterson sPeterson((checkFlag p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvM_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvM_trans_checkTurn:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvM sPeterson"
  and p2: "enablement_of  Peterson sPeterson((checkTurn p))"
shows " InvM (effects_of  Peterson sPeterson((checkTurn p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvM_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvM_trans_critical:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvM sPeterson"
  and p2: "enablement_of  Peterson sPeterson((critical p))"
shows " InvM (effects_of  Peterson sPeterson((critical p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvM_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvM_trans_release:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvM sPeterson"
  and p2: "enablement_of  Peterson sPeterson((release p))"
shows " InvM (effects_of  Peterson sPeterson((release p)))"
  apply (insert p0 p1 p2)
  apply (simp_all add:  InvM_def  Peterson_def  Peterson_enablement_def  Peterson_effect_def  Peterson_state.make_def)
  apply (auto ?)
done

theorem  InvM_trans1:
assumes
  p0: "reachable  Peterson sPeterson"
  and p1: " InvM sPeterson"
  and p2: "enablement_of  Peterson sPeterson aPeterson"
shows " InvM (effects_of  Peterson sPeterson aPeterson)"
  apply (insert p0 p1 p2)
  apply (cases  aPeterson)
  apply (simp add:  InvM_trans_trying)
  apply (simp add:  InvM_trans_setFlag)
  apply (simp add:  InvM_trans_setTurn)
  apply (simp add:  InvM_trans_checkFlag)
  apply (simp add:  InvM_trans_checkTurn)
  apply (simp add:  InvM_trans_critical)
  apply (simp add:  InvM_trans_release)
done

theorem  InvM_trans:
  "invariant_trans  Peterson InvM"
  apply (simp add: invariant_trans_def)
  apply (auto )
  apply (simp add:  InvM_trans1)
done

theorem  InvM_invariant:
  "invariant  Peterson InvM"
  apply (rule invariantI)
  apply (simp_all add:  InvM_start  InvM_trans)
done

theorem  InvM_reachable:
  "reachable  Peterson sPeterson ==>  InvM sPeterson"
  apply (insert  InvM_invariant)
  apply (auto intro: invariantReachable)
done
end
