BST(E): trait
  % Tree with binary search property
  assumes TotallyOrdered(E)

  introduces
    {}: 				-> BST[E]
    {__}:   		E		-> BST[E]
    insert: 		E, BST[E] 	-> BST[E]
    delete: 		E, BST[E]  	-> BST[E]
    isIn:   		E, BST[E]  	-> Bool
    leftChild:  	BST[E]       	-> BST[E]
    rightChild: 	BST[E]       	-> BST[E]
    parent:		BST[E]		-> BST[E] 
    successor:  	E, BST[E]	-> BST[E]
    predecessor:	E, BST[E]	-> BST[E] 
    minimum:    	BST[E] 		-> BST[E]
    maximum:    	BST[E]    	-> BST[E]
    root:       	BST[E]          -> E
    search:     	E, BST[E]       -> BST[E]
    hasLeft:    	BST[E]          -> Bool
    hasRight:   	BST[E]		-> Bool
    isLeaf:     	BST[E]		-> Bool
    preOrderWalk: 	BST[E]		-> Seq[E]
    inOrderWalk: 	BST[E]		-> Seq[E]
    postOrderWalk: 	BST[E]		-> Seq[E]

  asserts with e, n:E, t1, t2:BST[E]
    sort BST[E] generated by {}, insert;
    insert(e, {}) = {e};
    delete(e, {e}) = {};
    
    ~isIn(e, {});
    isIn(e, insert(n, t1) <=> e = n \/ isIn(e, t1));
  
    root(insert(e,t1)) = if (t1 = {}) then e
                                      else root(t1);

    search(n, {}) = {};
    search(n, insert(e,t1)) = if (t1 = {})
				then if (e = n) then insert(e,t1)
						else {}
				else
					if (n = root(t1))
						then t1
						else if (n < root(t1))
							then if (e <= root(t1))
								then (search(n,insert(e,leftChild(t1))))
								else (search(n,leftChild(t1)))
							else if (e > root(t1))
								then (search(n,insert(e,rightChild(t1))))
								else (search(n,rightChild(t1)));
		
    leftChild({}) = {};
    leftChild(insert(e,t1)) = if (t1 = {}) 
				then e
				else
					if (e <= root(t1)) then insert(e, leftChild(t1))
						           else leftChild(t1);

    rightChild({}) = {};
    rightChild(insert(e,t1)) = if (t1 = {})
				then e
				else 
					if (e > root(t1)) then insert(e, rightChild(t1))
                                        	          else rightChild(t1);

    ~hasLeft({});
    hasLeft(insert(e, t1)) = if (t1 = {}) then false
					  else if (e <= root(t1)) then true
                                                		  else (leftChild(t1) != {});

    ~hasRight({});
    hasRight(insert(e, t1)) = if (t1 = {})
				then e
				else
					if (e > root(t1)) then true
 						  	  else (rightChild(t1) != {});

    ~isLeaf({});
    isLeaf(insert(e,t1)) = if (t1 = {}) then true
					else false;


    parent({}) = {};
    parent(insert(e,t1)) = if (t1 = {}) then {}
				else parent(t1);

    minimum({}) = {};
    minimum(insert(e,t1)) = if (t1 = {}) 
				then e
				else	 
					if (e <= root(t1)) 
						then minimum(insert(e, leftChild(t1)))
						else if (hasLeft(t1))
							then minimum(leftChild(t1))
							else root(t1);

    maximum({}) = {};
    maximum(insert(e,t1)) = if (t1 = {})
				then e
				else
					if (e > root(t1))
						then maximum(insert(e,rightChild(t1)))
						else if (hasRight(t1))
							then maximum(rightChild(t1))
							else root(t1);

    preOrderWalk({}) = {};
    preOrderWalk(insert(e,t1)) = if (t1 = {})
					then Seq.empty().append(e)
					else 
						if (e <= root(t1))
				   			then  Seq.empty().append(root(t1)).append
							      (preOrderWalk(insert(e,leftChild(t1)))).append
							      (preOrderWalk(rightChild(t1)))
				   			else  Seq.empty().append(root(t1)).append
							      (preOrderWalk(leftChild(t1))).append
							      (preOrderWalk(insert(e,rightChild(t1))));

    inOrderWalk({}) = {};
    inOrderWalk(insert(e,t1)) = if (t1 = {})
					then Seq.empty().append(e)
					else 
						if (e <= root(t1))
				   			then  Seq.empty().append
							      (preOrderWalk(insert(e,leftChild(t1)))).append
							      (root(t1)).append
							      (preOrderWalk(rightChild(t1)))
				   			else  Seq.empty().append
							      (preOrderWalk(leftChild(t1))).append
							      (root(t1)).append
							      (preOrderWalk(insert(e,rightChild(t1))));

    postOrderWalk({}) = {};
    postOrderWalk(insert(e,t1)) = if (t1 = {})
					then Seq.empty().append(e)
					else 
						if (e <= root(t1))
				   			then  Seq.empty().append
							      (preOrderWalk(insert(e,leftChild(t1)))).append
							      (preOrderWalk(rightChild(t1))).append
							      (root(t1));
				   			else  Seq.empty().append
							      (preOrderWalk(leftChild(t1))).append
							      (preOrderWalk(insert(e,rightChild(t1)))).append
							      (root(t1));

    predecessor(n, {}) = {};
    predecessor(n, insert(e,t1)) = if (t1 = {} && e < n)
					then insert(e,t1)
					else
						if (e < n && e > predecessor(n,t1))
							then e
							else predecessor(n,t1); 

    successor(n, {}) = {};
    successor(n, insert(e,t1)) = if (t1 = {} && e > n)
					then insert(e,t1)
					else
						if (e > n && e < successor(n,t1))
							then e
							else successor(n,t1));
				
    implies i,j:Int, t1:BST[E], e,n:E
 
    \A t1 \A n (isIn(n, leftChild(t1)) => n <= root(t1));
    \A t1 \A n (isIn(n, rightChild(t1)) => n > root(t1));
    \A e \A n \A t1 (isIn(e, leftChild(t1)) && isIn(n, rightChild(t1)) => e < n);

    \A e \A t1 (isIn(e,t1) && (parent(e) = {}) => (root(t1) = e));

    \A e (hasLeft(e) && isIn(e,t1) => root(minimum(t1)) != e);
    \A e (hasRight(e) && isIn(e,t1) => root(maximum(t1)) != e);

    \E e \A n (minimum(t1) = e => (e <= n));
    \E e \A n (maximum(t1) = e => (e >= n));

    \A i \A j (0 <= i < j < length(inOrderWalk(t1)) <=> inOrderWalk(t1)[i] < inOrderWalk(t1)[j]);
    \A i (1 <= i < length(inOrderWalk(t1)) => inOrderWalk(t1)[i-1] = predecessor(inOrderWalk(t1)[i], t1);
    \A i (0 <= i < length(inOrderWalk(t1))-1 => inOrderWalk(t1)[i+1] = successor(inOrderWalk(t1)[i], t1);
    \A t1 (inOrderWalk(t1)[1] = minimum(t1));
    \A t1 (inOrderWalk(t1)[length(inOrderWalk(t1))-1] = maximum(t1));
 
    converts isIn, content, hasLeft, hasRight, isLeaf, preOrderWalk,
	inOrderWalk, postOrderWalk






