;;; ioa-mode.el --- IOA mode and a few commands. 

;;; COPYING
;;;
;;; Include copyright (left) information here. 

;;; AUTHORS
;;;
;;; Laura Dean (lgdean@mit.edu)
;;; Yogishwar Maharaj (ymaharaj@theory.lcs.mit.edu)
;;; Last updated by $Author: josh $ on $Date: 2003/03/13 17:36:16 $.

;;; README
;;;
;;; This package provides an attempt at an IOA emacs mode.

;;; INSTALL
;;;
;;; How to autoload IOA mode with every .ioa file:
;;;
;;; Put the following lines in your ~/.emacs file:
;;;
;;; (setq auto-mode-alist
;;;       (append '(("\\.ioa$" . ioa-mode))
;;;               auto-mode-alist))
;;;
;;; (setq load-path (cons "/path-to/IOA_Toolkit/" load-path))
;;; ; where "/path-to/IOA_Toolkit/" is replaced by the directory
;;; ; containing the ioa-mode.el file
;;;
;;; (autoload 'ioa-mode "ioa-mode" "IOA Mode." t)
;;;
;;;
;;; To turn font-lock (coloring) on for all IOA-mode buffers,
;;; use the following lines:
;;;
;;; (add-hook 'ioa-mode-hook
;;;           (lambda ()
;;;             (if window-system
;;;                 (turn-on-font-lock))))


;;; BUGS
;;;
;;; No online Documentation yet
;;;
;;; known bugs reported to me so far:
;;; font-locking doesn't work with XEmacs (have i fixed this? --LGD)


;;; TODO
;;;
;;; invariant, nonstandard stuff in Ramirez Thesis
;;; simulation blocks - simulation relations
;;; Set up some mode to "display" math characters - Unicode maybe?
;;;

;;;; CODE:

(defvar ioa-mode-hook nil
  "Normal hook run when entering IOA mode.")

(defvar ioa-mode-syntax-table nil
  "Syntax table used while in IOA mode.")

(if ioa-mode-syntax-table
    ()
  (setq ioa-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?% "<" ioa-mode-syntax-table)
  (modify-syntax-entry ?\n ">" ioa-mode-syntax-table)
  (modify-syntax-entry ?\f ">" ioa-mode-syntax-table)
  (modify-syntax-entry ?_ "_" ioa-mode-syntax-table)
  (modify-syntax-entry ?' "_" ioa-mode-syntax-table)
  (modify-syntax-entry ?\" ".   " ioa-mode-syntax-table)
  ;(modify-syntax-entry ?\\ "\\" ioa-mode-syntax-table)
  (modify-syntax-entry ?' "\'" ioa-mode-syntax-table))


(defvar ioa-mode-abbrev-table nil
  "Abbreviation table used while in IOA mode.")
(define-abbrev-table 'ioa-mode-abbrev-table ())


(defvar ioa-mode-map nil
  "Keymap for IOA mode.")


(if ioa-mode-map
    ()
  (setq ioa-mode-map (make-sparse-keymap))
  (define-key ioa-mode-map "\t" 'ioa-indent-line))

;;;; This should be written as follows, but it breaks
;;; (if ioa-mode-map   ;;; still to be done
;;;    nil
;;;  (let ((my-map (make-sparse-keymap)))
;;;    (define-key ioa-mode-map "\t" 'ioa-indent-line)
;;;    (setq ioa-mode-map my-map)))
	

(defun ioa-mode ()
  "Major mode for editing IOA code.
More documentation will be added here."
  (interactive)
  (kill-all-local-variables)
  (use-local-map ioa-mode-map)
  (setq local-abbrev-table ioa-mode-abbrev-table)
  (set-syntax-table ioa-mode-syntax-table)
  (setq mode-name "IOA")
  (setq major-mode 'ioa-mode)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'indent-tabs-mode)
  (setq indent-tabs-mode nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'ioa-indent-line)
  ;(make-local-variable 'indent-region-function)
  ;(setq indent-region-function 'ioa-indent-region)
  ;; Comment Variables and Defs
  (make-local-variable 'comment-start)
  (setq comment-start "%")
  (make-local-variable 'comment-end)
  (setq comment-end "")  ;; isn't this already there by default?
  ;(make-local-variable 'comment-start-skip)
  (make-local-variable 'comment-column)
  (setq comment-column 32)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'ioa-comment-indent)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(ioa-font-lock-keywords))
  ;(make-local-variable ')
  (run-hooks 'ioa-mode-hook))


;;; Font Locking and Keyword Highlighting

;;; font-lock stuff
;(font-lock-add-keywords
; 'ioa-mode
; (list (concat "\\<\\("  ;;later, use eval-when-compile regexp
;               "if\\|then\\|else\\|elseif\\|fi\\|do\\|od\\|for\\)\\>")))
;(font-lock-add-keywords
; 'ioa-mode
; '("\\<\\(input\\|output\\|internal\\)\\>"))
;(font-lock-add-keywords
; 'ioa-mode
; '("\\<\\(enumeration\\|union\\|tuple\\) of\\>"))
;(font-lock-add-keywords
; 'ioa-mode
; '("\\<\\(uses\\|assumes\\)\\>"))
;(font-lock-add-keywords
; 'ioa-mode
; '("\\<\\(simulation\\|backward\\|forward\\|from\\|to\\)\\>"))



(defvar ioa-font-lock-keywords
  (let ((wordlist (mapconcat 'identity
                             '("automaton\\|signature\\|states\\|transitions"
                               "schedule\\|initially\\|proof\\|tasks"
                               "choose\\|local\\|pre\\|eff\\|where"
                               "if\\|then\\|else\\|elseif\\|fi"
			       "do\\|od\\|for\\|in\\|ensuring"
                               "while\\|fire\\|yield\\|det"
                               "input\\|output\\|internal"
                               "\\(enumeration\\|union\\|tuple\\) of"
                               "type\\|uses\\|axioms\\|assumes"
                               "simulation\\|backward\\|forward\\|from\\|to"
			       "components\\|hidden\\|start\\|case"
                               )
                             "\\|")))
    (list
     ;; keywords
     (cons (concat "\\<\\(" wordlist "\\)\\>") 1)
;     ;; type decl stuff
;     '("\\btype[ \t]+\\([a-zA-Z_]+[a-zA-Z0-9_]*\\)"
;       1 font-lock-type-face)
     ;; for comments.  grr, i liked font-lock-add-keywords
     ;; because it just worked.  damn xemacs.
     '("%.*$"
       0 font-lock-comment-face)
     ))
  "Words for font-lock to highlight in IOA mode.")


;;; Indentation and Tabbing

;; some useful REs
(defvar ioa-indent-level nil
  "Number of spaces to indent lines of IOA code.")

(if ioa-indent-level () (setq ioa-indent-level 2))

(defvar ioa-indent-starter
  "\\(\\(choose\\|pre\\|eff\\|components\\|hidden\\|.* then\\|.* do\\|signature\\|states\\|transitions\\|tasks\\|initially\\|ensuring\\start\\|)[ \t]*$\\)"
  "r.e. for words that suggest the following code should be indented")

(defvar ioa-indent-keeper
  "\\(\\(choose\\|pre\\|eff\\|components\\|initially\\|ensuring\\|hidden\\|start\\|.* then\\|.* do\\)[ \t]*\\w.*$\\)"
  "r.e. for lines that suggest the following code should be at the same level")

(defvar ioa-transition-type
  "\\(input\\|output\\|internal\\)$"
  "r.e. for input, output, or internal (not followed by more text)")

(defvar ioa-transition-head
  "\\(input\\|output\\|internal\\)\\W+\\w+.*$"
  "r.e. for input, output, or internal, followed by more text")

(defvar ioa-indent-ender
  "\\(\\(fi\\|od\\)[ \t]*$\\)"
  "r.e. for words that suggest following code be less indented")

(defvar ioa-middle-of-list
  "\\(.*\\([,;]\\|/\\\\\\|\\\\/\\)[ \t]*$\\)"
  "r.e. for lines ending in /\\ or  ; or ,")
;(yes, i know that some lines ending in , are not declarations,
;but this works for what I'm tring to do right now.)")

;(defvar ioa-within-if 0
;  "tells whether we're within an if, and if so, how many")
;
;(defvar ioa-within-do 0
;  "tells whether we're within a do...od, and if so, how many")

(defvar ioa-typeinfo
  "\\([ \t]*\\(type\\|uses\\|axioms\\|assumes\\|input\\|output\\|internal\\)\\)"
  "r.e. for type, assumes, or uses and axioms at beginning")


;; actual functions for indenting
(defun ioa-comment-indent ()  ;; was lsl-
  ;; my question is, when is this function actually called?
  (skip-chars-backward " \t")
  ;; if the line is blank, put the comment at the beginning,
  ;; else at comment-column
  (if (bolp) 0 (max (1+ (current-column)) comment-column)))

(defun ioa-indent-line ()
  "Indent current line of IOA code."
  (interactive)
  (let ((c 0))
    (save-excursion
      (forward-line -1)
      (back-to-indentation)
      (while (and (eolp) (not (bobp))) ;; bad if line 1 has just whitespace
        (forward-line -1)
        (back-to-indentation))
      (setq c
            (cond ((eolp) 0)
                  ((looking-at "\\(%\\)")  ;bleh
                   (current-column))
                  ((or (looking-at "\\(automaton\\)")
                       (looking-at "\\(schedule\\)"))
                   (+ (current-column) ioa-indent-level))
                  ((looking-at ioa-indent-starter)
                   (+ (current-column) ioa-indent-level))
                  ((looking-at ioa-indent-keeper)
                   (current-column))
                  ((looking-at ioa-transition-type)
                   (+ (current-column) ioa-indent-level))
                  ((looking-at ioa-transition-head)
                   (+ (current-column) ioa-indent-level)) 
                  ((looking-at ioa-typeinfo)
                   (current-column))
                  ((not (looking-at ioa-middle-of-list))
                   (- (current-column) ioa-indent-level))
                  (t (current-column)))))
    (indent-line-to c))                 ;indent line to this level
  )

(provide 'ioa-mode)

;; for later:
;; ioa-fill-paragraph






;;; ioa-mode.el ends here







