;;; An attempt at an LP mode ripped off from Laura Dean's IOA mode

;;; Toh Ne Win (tohn@mit.edu)
;;; Last updated by $Author: tohn $ on $Date: 2002/03/06 23:46:37 $.

;; How to autoload LP mode with every .lp file:
;;
;; Put the following lines in your ~/.emacs file:
;;
;; (setq auto-mode-alist
;;       (append '(("\\.lp$" . lp-mode))
;;               auto-mode-alist))
;;
;; (setq load-path (cons "/path-to/LP_Toolkit/" load-path))
;; ; where "/path-to/LP_Toolkit/" is replaced by the directory
;; ; containing the lp-mode.el file
;;
;; (autoload 'lp-mode "lp-mode" "LP Mode." t)
;;
;;
;; To turn font-lock (coloring) on for all lp-mode buffers,
;; use the following lines:
;;
;; (add-hook 'lp-mode-hook
;;           (lambda ()
;;             (if window-system
;;                 (turn-on-font-lock))))

;; known bugs reported to me so far:
;; font-locking doesn't work with XEmacs (have i fixed this? --LGD)

(defvar lp-mode-hook nil
  "Normal hook run when entering LP mode.")

(defvar lp-mode-syntax-table nil
  "Syntax table used while in LP mode.")

(defvar lp-mode-abbrev-table nil
  "Abbrev table used while in LP mode.")
(define-abbrev-table 'lp-mode-abbrev-table ())

(if lp-mode-syntax-table
    ()
  (setq lp-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?% "<" lp-mode-syntax-table)
  (modify-syntax-entry ?\n ">" lp-mode-syntax-table)
  (modify-syntax-entry ?\f ">" lp-mode-syntax-table)
  (modify-syntax-entry ?_ "_" lp-mode-syntax-table)
  (modify-syntax-entry ?' "_" lp-mode-syntax-table)
  (modify-syntax-entry ?\" ".   " lp-mode-syntax-table)
  ;(modify-syntax-entry ?\\ "\\" lp-mode-syntax-table)
  (modify-syntax-entry ?' "\'" lp-mode-syntax-table))

(defvar lp-mode-map nil
  "Keymap for LP mode.")

(if lp-mode-map   ;;; still to be done
    ()
  (setq lp-mode-map (make-sparse-keymap))
  (define-key lp-mode-map "\t" 'lp-indent-line)
  (define-key lp-mode-map "]" 'lp-indent-and-insert)
  )

(defun lp-indent-and-insert ()
  (interactive)
  (insert "]")
  (lp-indent-line))

(defun lp-mode ()
  "Major mode for editing LP code."
  (interactive)
  (kill-all-local-variables)
  (use-local-map lp-mode-map)
  (setq local-abbrev-table lp-mode-abbrev-table)
  (set-syntax-table lp-mode-syntax-table)
  (setq mode-name "LP")
  (setq major-mode 'lp-mode)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'indent-tabs-mode)
  (setq indent-tabs-mode nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'lp-indent-line)
  (make-local-variable 'comment-start)
  (setq comment-start "%")
  (make-local-variable 'comment-end)
  (setq comment-end "")  ;; isn't this already there by default?
  (make-local-variable 'comment-column)
  (setq comment-column 32)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function (lambda () comment-column)) ;'lp-comment-indent)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(lp-font-lock-keywords))
  ;(make-local-variable ')
  (run-hooks 'lp-mode-hook))

(defun lp-make-word-list (words)
  "Make regular expresion union of a list of strings"
  (concat "\\<\\(" 
	  (mapconcat 'identity
		     words
		     "\\|")
	  "\\)\\>"))


(defvar lp-font-lock-keywords
  (let (
	(keyword-list (lp-make-word-list
		   '("in" "by" "with" "immune" "cases" "contradiction" "spec"
		     "specializing" 
		     )))
	(name-list (lp-make-word-list
		   '("conjecture" "con"
		     )))
	(command-list (lp-make-word-list
		   '("clear" "thaw" "display" "declare" "prove" "resume" "instantiate" "assert"
		     "di" "disp" "decl" "pr" "res" "inst" "qed" "complete" "ass"
		     "make" "set"
		     )))
	(function-list (lp-make-word-list
		   '("con-op" "con-var" "contains-op" "contains-var"
		     )))
	)
    (list
     ;; keywords
     '("\\(%\\|\\[\\]\\|<>\\).*$"
       0 font-lock-comment-face)
     (list keyword-list 1 'font-lock-keyword-face)
     (list name-list 1 'font-lock-variable-name-face)
     (list command-list 1 'font-lock-function-name-face)
     (list function-list 1 'font-lock-builtin-face)
     ))
  "Words for font-lock to highlight in LP mode.")

;;; indent stuff

(defvar lp-indent-level 2
  "Number of spaces to indent lines of LP code.")

(if lp-indent-level () (setq lp-indent-level 2))

(defvar lp-indent-starter
  "\\(<>\\).*"
  "r.e. for words that suggest the following code should be indented")


(defvar lp-indent-ender
  "\\(\\[\\]\\).*"
  "r.e. for words that suggest following code be less indented")



;(defvar lp-within-if 0
;  "tells whether we're within an if, and if so, how many")
;
;(defvar lp-within-do 0
;  "tells whether we're within a do...od, and if so, how many")

(defvar lp-typeinfo
  "\\([ \t]*\\(type\\|uses\\|assumes\\|input\\|output\\|internal\\)\\)"
  "r.e. for type, assumes, or uses at beginning")

;; actual functions for indenting
(defun lp-comment-indent ()  ;; was lsl-
  ;; my question is, when is this function actually called?
  (skip-chars-backward " \t")
  ;; if the line is blank, put the comment at the beginning,
  ;; else at comment-column
  (if (bolp) 0 (max (1+ (current-column)) comment-column)))

(defun lp-indent-line ()
  "Indent current line of LP code."
  (interactive)
  (let ((c -1)
	(backwards-indent ())
	)
    (save-excursion
      (back-to-indentation)
      (if (looking-at lp-indent-ender)
	  (setq backwards-indent t)
	))
    (if (= c -1)
	(save-excursion
	  (forward-line -1)
	  (back-to-indentation)
	  (while (and (eolp) (not (bobp))) 
	    (forward-line -1)
	    (back-to-indentation))
	  (setq c
		(cond ((eolp) 0)
		      ((and backwards-indent (looking-at lp-indent-starter)) (current-column))
		      (backwards-indent (- (current-column) lp-indent-level))
		      ((looking-at lp-indent-starter)
		       (+ (current-column) lp-indent-level ))
		      (t (current-column))
		      )
		)))
    (if (> c -1) 
	(if (>= (current-column) c) (save-excursion (indent-line-to c))
	  (indent-line-to c))
      )))

(provide 'lp-mode)

;;; for later:
; lp-fill-paragraph

